#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include "malem-nautilus-client.h"

#include <libnautilus-extension/nautilus-extension-types.h>
#include <libnautilus-extension/nautilus-file-info.h>
#include <libnautilus-extension/nautilus-info-provider.h>

#include <libgnomevfs/gnome-vfs-utils.h>
#include <string.h>

#define TASK_DATA_KEY "malem-nautilus-client-task"

static GObjectClass *parent_class;
static GType the_type = G_TYPE_INVALID;
 
static void
on_client_file_weak_notify (MalemNautilusClient *client,
			    NautilusFileInfo    *file)
{
	client->files = g_list_remove (client->files, file);
}

static gint
task_find_func (MalemFileTransfer *transfer,
		NautilusFileInfo  *file)
{
	gchar *uri = nautilus_file_info_get_uri (file);
	MalemFileTransferDirection direction;
	
	direction = malem_file_transfer_get_file_direction (transfer, uri);
	return !(direction != 0);
}

static NautilusOperationResult
impl_update_file_info (NautilusInfoProvider     *provider,
		       NautilusFileInfo         *file,
		       GClosure                 *update_complete,
		       NautilusOperationHandle **handle)
{
	MalemNautilusClient *client = MALEM_NAUTILUS_CLIENT (provider);
	GList *llink;
	MalemTask *task;
	gint progress;
	gchar emblem[16];

	llink = g_list_find (client->files, file);

	if (llink == NULL) {
		client->files = g_list_prepend (client->files, file);

		g_object_weak_ref (G_OBJECT (file),
				   (GWeakNotify) on_client_file_weak_notify,
				   client);
	}
	
	task = MALEM_TASK (g_object_get_data (G_OBJECT (file), TASK_DATA_KEY));
	
	if (task == NULL) {
		llink = g_list_find_custom (client->tasks,
					    file,
					    (GCompareFunc) task_find_func);
		if (llink == NULL)
			/* file is not involved in any task */
			return NAUTILUS_OPERATION_COMPLETE;

		task = MALEM_TASK (llink->data);
		g_object_ref (task);
		g_object_set_data_full (G_OBJECT (file), TASK_DATA_KEY,
					task, (GDestroyNotify) g_object_unref);
	}

	progress = malem_task_get_progress (task);
	g_snprintf (emblem, G_N_ELEMENTS (emblem),
		    "progress-%03d", progress - progress % 25);
	nautilus_file_info_add_emblem (file, emblem);

  	return NAUTILUS_OPERATION_COMPLETE;
}

static void
file_invalidate_func (NautilusFileInfo  *file,
		      MalemTask         *task)
{
	gchar *uri = nautilus_file_info_get_uri (file);
	MalemTaskStatus status = malem_task_get_status (task);
	
	if (g_object_get_data (G_OBJECT (file), TASK_DATA_KEY) == (gpointer) task) {
		if (status == MALEM_TASK_STATUS_COMPLETED ||
		    status == MALEM_TASK_STATUS_ABORTED) {
			g_object_steal_data (G_OBJECT (file), TASK_DATA_KEY);
			g_object_unref (task);
		}
		nautilus_file_info_invalidate_extension_info (file);
	}
}

static void
invalidate_file_for_task (MalemNautilusClient *client,
			  MalemTask	      *task)
{
	g_list_foreach (client->files, (GFunc) file_invalidate_func, task);
}

static void
on_task_progress_changed (MalemTask           *task,
			  guint                progress,
			  MalemNautilusClient *client)
{
	invalidate_file_for_task (client, task);
}

static void
on_task_status_changed (MalemTask           *task,
			MalemTaskStatus      status,
			MalemNautilusClient *client)
{
	if (status == MALEM_TASK_STATUS_COMPLETED ||
	    status == MALEM_TASK_STATUS_ABORTED) {
		invalidate_file_for_task (client, task);
		client->tasks = g_list_remove (client->tasks, task);
		g_object_unref (client->tasks);
	}
}

static void
on_task_file_added (MalemTask           *task,
		    const gchar         *filename,
		    MalemNautilusClient *client)
{
	GList *llink;
	NautilusFileInfo *file;
	
	for (llink = client->files; llink != NULL; llink = llink->next) {
		file = NAUTILUS_FILE_INFO (llink->data);
		if (strcmp (nautilus_file_info_get_uri (file), filename) == 0) {
			nautilus_file_info_invalidate_extension_info (file);
			g_object_ref (task);
			g_object_set_data_full (G_OBJECT (file), TASK_DATA_KEY,
						task, (GDestroyNotify) g_object_unref);
			return;
		}
	}
}

static void
register_task (MalemTask           *task,
	       MalemNautilusClient *client)
{
	if (MALEM_IS_FILE_TRANSFER (task)) {
		client->tasks = g_list_prepend (client->tasks, task);
		g_object_ref (task);
		g_signal_connect (task, "progress-changed",
				  (GCallback) on_task_progress_changed, client);
		g_signal_connect (task, "status-changed",
				  (GCallback) on_task_status_changed, client);
		g_signal_connect (task, "file-added",
				  (GCallback) on_task_file_added, client);
	}
}

static void
on_task_registered (MalemFactory        *factory,
		    MalemTask           *task,
		    gchar              **interfaces,
		    MalemNautilusClient *client)
{
	register_task (task, client);
}

static void
malem_nautilus_client_init (MalemNautilusClient *client)
{
	MalemTask **tasks;
	gint i;

	client->tasks = NULL;
	client->files = NULL;

	client->factory = MALEM_FACTORY (malem_factory_client_get_default ());
	g_object_ref (client->factory);

	client->task_registered_handler_id =
		g_signal_connect (G_OBJECT (client->factory), "task-registered",
				  G_CALLBACK (on_task_registered), client);
	
	tasks = malem_factory_list_tasks_for_interface (client->factory,
							"org.gnome.Mathusalem.FileTransfer");
	for (i = 0; tasks[i] != NULL; i++) {
		register_task (tasks[i], client);
		g_object_unref (tasks[i]);
	}
	g_free (tasks);
}

static void
malem_nautilus_client_finalize (GObject *object)
{
	MalemNautilusClient *client = MALEM_NAUTILUS_CLIENT (object);
	GList *llink;

	for (llink = client->tasks; llink != NULL; llink = llink->next)
		g_object_unref (llink->data);
	g_list_free (client->tasks);
	g_list_free (client->files);

	g_object_unref (client->factory);
}

static void
malem_nautilus_client_class_init (MalemNautilusClientClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);
	
	object_class->finalize = malem_nautilus_client_finalize;
}

static void 
nautilus_info_provider_iface_init (NautilusInfoProviderIface *iface)
{
	iface->update_file_info = impl_update_file_info;
}

GType
malem_nautilus_client_get_type (void) 
{
	return the_type;
}

void
malem_nautilus_client_register_type (GTypeModule *module)
{
	
	static const GTypeInfo type_info = {
		sizeof (MalemNautilusClientClass),
		(GBaseInitFunc) NULL,
		(GBaseFinalizeFunc) NULL,
		(GClassInitFunc) malem_nautilus_client_class_init,
		NULL, 
		NULL,
		sizeof (MalemNautilusClient),
		0,
		(GInstanceInitFunc) malem_nautilus_client_init,
	};

	the_type = g_type_module_register_type (module,
						G_TYPE_OBJECT,
						"MalemNautilusClient",
						&type_info, 
						0);

	static const GInterfaceInfo iface_info = {
		(GInterfaceInitFunc) nautilus_info_provider_iface_init,
		NULL,
		NULL,
	};

	g_type_module_add_interface (module,
				     the_type,
				     NAUTILUS_TYPE_INFO_PROVIDER,
				     &iface_info);
}
