/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <dbus/dbus-glib.h>
#include <gtk/gtk.h>

#include "malem-factory-client.h"
#include "malem-task-client.h"
#include "malem-factory.h"
#include "malem-marshal.h"

#define MALEM_FACTORY_CLIENT_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o),	\
					     MALEM_TYPE_FACTORY_CLIENT,		\
					     MalemFactoryClientPrivate))

struct _MalemFactoryClientPrivate {
	DBusGProxy      *proxy;
};

static void		 malem_factory_iface_init		(MalemFactoryIface *iface);
static MalemTask	*malem_factory_client_register_task	(MalemFactory  *factory,
								 const gchar  **interfaces,
								 const gchar   *title,
								 const gchar   *owner);
static MalemTask	**malem_factory_client_list_tasks	(MalemFactory  *factory);
static MalemTask	**malem_factory_client_list_tasks_for_interface
								(MalemFactory  *factory,
								 const gchar   *interface);
static void		 malem_factory_client_get_version	(MalemFactory  *factory,
								 gint          *major,
								 gint          *minor,
								 gint          *micro);

G_DEFINE_TYPE_WITH_CODE (MalemFactoryClient, malem_factory_client, G_TYPE_OBJECT,
			 G_IMPLEMENT_INTERFACE (MALEM_TYPE_FACTORY,
				 		malem_factory_iface_init))

static void
on_dbus_task_registered (DBusGProxy          *proxy,
			 const gchar         *object_path,
			 const gchar        **interfaces,
			 MalemFactoryClient  *factory)
{
	MalemTask *task = MALEM_TASK (g_object_new (malem_task_client_get_type (interfaces),
						    "object-path", object_path, 
						    NULL));
	malem_factory_task_registered (MALEM_FACTORY (factory), task, interfaces);
	g_object_unref (task);
}
		 		
static GObject *
malem_factory_client_constructor (GType                  type,
				  guint                  n_construct_properties,
				  GObjectConstructParam *construct_properties)
{
        MalemFactoryClient *factory;
	DBusGConnection *connection;
        GError *error = NULL;

        factory = MALEM_FACTORY_CLIENT (G_OBJECT_CLASS (malem_factory_client_parent_class)->constructor
                              (type, n_construct_properties, construct_properties));

	connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);

	if (connection == NULL) {
		g_printerr ("Failed to open connection to bus: %s\n",
                	    error->message);
		g_error_free (error);
		factory->priv->proxy = NULL;
		return G_OBJECT (factory);
	}
	
	factory->priv->proxy = dbus_g_proxy_new_for_name (connection,
							  "org.gnome.Mathusalem",
							  "/org/gnome/Mathusalem",
							  "org.gnome.Mathusalem");

	dbus_g_object_register_marshaller ((GClosureMarshal) malem_marshal_VOID__STRING_POINTER,
					   G_TYPE_NONE,
					   DBUS_TYPE_G_OBJECT_PATH,
					   G_TYPE_STRV,
					   G_TYPE_INVALID);

	dbus_g_proxy_add_signal (factory->priv->proxy,
				 "TaskRegistered",
				 DBUS_TYPE_G_OBJECT_PATH,
				 G_TYPE_STRV,
				 G_TYPE_INVALID);

	dbus_g_proxy_connect_signal (factory->priv->proxy,
				     "TaskRegistered",
				     G_CALLBACK (on_dbus_task_registered),
				     factory, NULL);

	return G_OBJECT (factory);
}

static void
malem_factory_client_init (MalemFactoryClient *factory_client)
{
	factory_client->priv = MALEM_FACTORY_CLIENT_GET_PRIVATE (factory_client);
}

static void
malem_factory_client_finalize (GObject *object)
{
	MalemFactoryClient *factory = MALEM_FACTORY_CLIENT (object);

	g_object_unref (factory->priv->proxy);

	G_OBJECT_CLASS (malem_factory_client_parent_class)->finalize (object);
}

static void
malem_factory_iface_init (MalemFactoryIface *iface)
{
	iface->register_task = malem_factory_client_register_task;
	iface->list_tasks = malem_factory_client_list_tasks;
	iface->list_tasks_for_interface = malem_factory_client_list_tasks_for_interface;
	iface->get_version = malem_factory_client_get_version;
}

static void
malem_factory_client_class_init (MalemFactoryClientClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	
	object_class->constructor = malem_factory_client_constructor;
	object_class->finalize = malem_factory_client_finalize;

	g_type_class_add_private (klass, sizeof (MalemFactoryClientPrivate));
}

MalemFactoryClient *
malem_factory_client_get_default (void)
{
	static MalemFactoryClient *factory = NULL;

	if (G_UNLIKELY (factory == NULL)) {
		factory = MALEM_FACTORY_CLIENT (g_object_new (MALEM_TYPE_FACTORY_CLIENT, NULL));
		g_object_add_weak_pointer (G_OBJECT (factory), (gpointer) &factory);
	}

	return factory;
}

static MalemTask *
malem_factory_client_register_task (MalemFactory  *factory,
				    const gchar  **interfaces,
				    const gchar   *title,
				    const gchar   *owner)
{
	gchar *object_path;
	GError *error = NULL;
	MalemFactoryClient *factory_client = MALEM_FACTORY_CLIENT (factory);
	MalemTask *task;

	if (!dbus_g_proxy_call (factory_client->priv->proxy, "RegisterTask", 
				&error,
				G_TYPE_STRV, interfaces,
				G_TYPE_STRING, title,
				G_TYPE_STRING, owner,
				G_TYPE_INVALID,
				DBUS_TYPE_G_OBJECT_PATH, &object_path,
				G_TYPE_INVALID)) {
		g_printerr ("Error while registering task: %s\n", 
			    error->message);
		g_error_free (error);
		return NULL;
	}

	task = MALEM_TASK (g_object_new (malem_task_client_get_type (interfaces), 
					 "object-path", object_path, 
					 NULL));

	g_free (object_path);
	malem_factory_task_registered (MALEM_FACTORY (factory), task, interfaces);

	return task;
}

static MalemTask **
get_tasks_from_paths (MalemFactoryClient  *client,
		      gchar              **paths)
{
	DBusGProxy *proxy;
	GPtrArray *array;
	gchar **interfaces;
	gpointer task;
	gint i;
	GError *error = NULL;
	
	array = g_ptr_array_new ();
	
	for (i = 0; paths[i] != NULL; i++) {
		proxy = dbus_g_proxy_new_from_proxy (client->priv->proxy,
						     "org.gnome.Mathusalem.Task",
						     paths[i]);

		if (!dbus_g_proxy_call (proxy, "ListInterfaces",
					&error,
					G_TYPE_INVALID,
					G_TYPE_STRV, &interfaces,
					G_TYPE_INVALID)) {
			g_printerr ("Error while listing task interfaces: %s\n", 
				    error->message);
			g_error_free (error);
			continue;
		}
		
		task = g_object_new (malem_task_client_get_type ((const gchar **)interfaces),
				     "object-path", paths[i], NULL);
		
		g_ptr_array_add (array, task);
		g_strfreev (interfaces);
	}
	
	g_ptr_array_add (array, NULL);
	return (MalemTask **) g_ptr_array_free (array, FALSE);
}

static MalemTask **
malem_factory_client_list_tasks (MalemFactory  *factory)
{
	MalemFactoryClient *factory_client = MALEM_FACTORY_CLIENT (factory);
	GError *error = NULL;
	gchar **paths;
	MalemTask **tasks;
	
	if (!dbus_g_proxy_call (factory_client->priv->proxy, "ListTasks",
				&error,
				G_TYPE_INVALID,
				G_TYPE_STRV, &paths,
				G_TYPE_INVALID)) {
		g_warning ("Error while listing task\n");
		g_error_free (error);
		return NULL;
	}
	
	tasks = get_tasks_from_paths (factory_client, paths);
	g_strfreev (paths);
	return tasks;
}

static MalemTask **
malem_factory_client_list_tasks_for_interface (MalemFactory  *factory,
					       const gchar   *interface)
{
	MalemFactoryClient *factory_client = MALEM_FACTORY_CLIENT (factory);
	GError *error = NULL;
	gchar **paths;
	MalemTask **tasks;

	if (!dbus_g_proxy_call (factory_client->priv->proxy, "ListTasksForInterface",
				&error,
				G_TYPE_STRING, interface,
				G_TYPE_INVALID,
				G_TYPE_STRV, &paths,
				G_TYPE_INVALID)) {
		g_warning ("Error while listing task\n");
		g_error_free (error);
		return NULL;
	}

	tasks = get_tasks_from_paths (factory_client, paths);
	g_strfreev (paths);
	return tasks;
}

static void
malem_factory_client_get_version (MalemFactory  *factory,
				  gint          *major,
				  gint          *minor,
				  gint          *micro)
{
	MalemFactoryClient *factory_client = MALEM_FACTORY_CLIENT (factory);
	GError *error;
	GArray *array;
	gint *version;

	if (!dbus_g_proxy_call (factory_client->priv->proxy, "RegisterTask", 
				&error,
				G_TYPE_INVALID,
				DBUS_TYPE_G_INT_ARRAY, &array,
				G_TYPE_INVALID)) {
		g_warning ("Error while registering task\n");
		g_error_free (error);
		return;
	}

	version = (gint *) array->data;
	*major = version[0];
	*minor = version[1];
	*micro = version[2];
	
	g_array_free (array, TRUE);
}
