/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <dbus/dbus-glib.h>

#include "malem-task-client.h"
#include "malem-task.h"
#include "malem-interfaces.h"

#define MALEM_TASK_CLIENT_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o),	\
					  MALEM_TYPE_TASK_CLIENT,		\
					  MalemTaskClientPrivate))

struct _MalemTaskClientPrivate {
	gboolean	  has_status;
	MalemTaskStatus   status;
	gboolean	  has_progress;
	guint             progress;
	gchar            *title;
	gchar            *owner;
};

enum {
	PROP_0,
	PROP_OBJECT_PATH,
	PROP_STATUS,
	PROP_TITLE,
	PROP_PROGRESS,
	PROP_OWNER
};

static gpointer malem_task_client_parent_class = NULL;

static void
on_dbus_status_changed (DBusGProxy      *proxy,
			guint            status,
			MalemTaskClient *task)
{
	task->priv->status = status;
	task->priv->has_status = TRUE;
	malem_task_status_changed (MALEM_TASK (task), status);
	if (status == MALEM_TASK_STATUS_COMPLETED || 
	    status == MALEM_TASK_STATUS_ABORTED) {
		g_object_unref (task->proxy);
		task->proxy = NULL;
	}
}

static void
on_dbus_progress_changed (DBusGProxy      *proxy,
			  guint            progress,
			  MalemTaskClient *task)
{
	task->priv->progress = progress;
	task->priv->has_progress = TRUE;
	malem_task_progress_changed (MALEM_TASK (task), progress);
}

static void
on_dbus_action_triggered (DBusGProxy      *proxy,
			  const gchar     *action_name,
			  MalemTask       *task)
{
	malem_task_action_triggered (task, action_name);
}


static GObject *
malem_task_client_constructor (GType                  type,
			       guint                  n_construct_properties,
			       GObjectConstructParam *construct_properties)
{
        MalemTaskClient *task;
	DBusGConnection *connection;
        GError *error = NULL;

        task = MALEM_TASK_CLIENT (G_OBJECT_CLASS (malem_task_client_parent_class)->constructor
                              (type, n_construct_properties, construct_properties));
	
	connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);

	if (connection == NULL) {
		g_printerr ("Failed to open connection to bus: %s\n",
                	    error->message);
		g_error_free (error);
		task->proxy = NULL;
		return G_OBJECT (task);
	}
	
	task->proxy = dbus_g_proxy_new_for_name (connection,
						       "org.gnome.Mathusalem",
						       task->object_path,
						       "org.gnome.Mathusalem.Task");

	dbus_g_proxy_add_signal (task->proxy,
				 "StatusChanged",
				 G_TYPE_UINT,
				 G_TYPE_INVALID);

	dbus_g_proxy_add_signal (task->proxy,
				 "ProgressChanged",
				 G_TYPE_UINT,
				 G_TYPE_INVALID);
	dbus_g_proxy_add_signal (task->proxy,
				 "ActionTriggered",
				 G_TYPE_STRING,
				 G_TYPE_INVALID);


	dbus_g_proxy_connect_signal (task->proxy,
				     "StatusChanged",
				     G_CALLBACK (on_dbus_status_changed),
				     task, NULL);

	dbus_g_proxy_connect_signal (task->proxy,
				     "ProgressChanged",
				     G_CALLBACK (on_dbus_progress_changed),
				     task, NULL);

	dbus_g_proxy_connect_signal (task->proxy,
				     "ActionTriggered",
				     G_CALLBACK (on_dbus_action_triggered),
				     task, NULL);

	/* doing it here allows us to assume the proxy is created when the
	   iface instances are initialized */
	malem_object_init_interfaces (G_OBJECT (task));

	return G_OBJECT (task);
}

static void
malem_task_client_init (MalemTaskClient *task)
{
	task->priv = MALEM_TASK_CLIENT_GET_PRIVATE (task);
}

static void
malem_task_client_finalize (GObject *object)
{
	MalemTaskClient *task = MALEM_TASK_CLIENT (object);

	g_free (task->object_path);
	g_free (task->priv->title);
	g_free (task->priv->owner);
	if (task->proxy != NULL)
		g_object_unref (task->proxy);

	/* chainup */
	G_OBJECT_CLASS (malem_task_client_parent_class)->finalize (object);
}

static MalemTaskStatus
malem_task_client_get_status (MalemTaskClient *task)
{
	GError *error = NULL;
	if (!task->priv->has_status && task->proxy != NULL) {
		if (!dbus_g_proxy_call (task->proxy, "GetStatus", &error,
					G_TYPE_INVALID,
					G_TYPE_UINT, &task->priv->status,
					G_TYPE_INVALID)) {
			g_warning ("Could not get the task status of %s\n", 
				   task->object_path);
			return 0;
		}
		task->priv->has_status = TRUE;
	}
	return task->priv->status;
}

static const gchar *
malem_task_client_get_title (MalemTaskClient *task)
{
	GError *error = NULL;
	if (!task->priv->title &&
	    task->proxy != NULL &&
	    !dbus_g_proxy_call (task->proxy, "GetStatus", &error,
				G_TYPE_INVALID,
				G_TYPE_STRING, &task->priv->title,
				G_TYPE_INVALID)) {
		g_warning ("Could not get the task title of %s\n", 
			   task->object_path);
		return NULL;
	}
	return task->priv->title;
}

static guint
malem_task_client_get_progress (MalemTaskClient *task)
{
	GError *error = NULL;
	if (!task->priv->has_progress && task->proxy != NULL) {
		if (!dbus_g_proxy_call (task->proxy, "GetStatus", &error,
					G_TYPE_INVALID,
					G_TYPE_UINT, &task->priv->progress,
					G_TYPE_INVALID)) {
			g_warning ("Could not get the task progress of %s\n", 
				   task->object_path);
			return 0;
		}
		task->priv->has_progress = TRUE;
	}
	return task->priv->progress;
}

static const gchar *
malem_task_client_get_owner (MalemTaskClient *task)
{
	GError *error = NULL;
	if (!task->priv->owner && task->proxy != NULL &&
	    !dbus_g_proxy_call (task->proxy, "GetStatus", &error,
				G_TYPE_INVALID,
				G_TYPE_STRING, &task->priv->owner,
				G_TYPE_INVALID)) {
		g_warning ("Could not get the task owner of %s\n", 
			   task->object_path);
		return NULL;
	}
	return task->priv->owner;
}

static void
malem_task_client_get_property (GObject    *object,
				guint       prop_id,
				GValue     *value,
				GParamSpec *pspec)
{
	MalemTaskClient *task = MALEM_TASK_CLIENT (object);

	switch (prop_id)
	{
		case PROP_OBJECT_PATH:
			g_value_set_string (value, task->object_path);
			break;
		case PROP_STATUS:
			g_value_set_enum (value, 
					  malem_task_client_get_status (task));
			break;
		case PROP_TITLE:
			g_value_set_string (value, 
					    malem_task_client_get_title (task));
			break;
		case PROP_PROGRESS:
			g_value_set_uint (value,
					  malem_task_client_get_progress (task));
			break;
		case PROP_OWNER:
			g_value_set_string (value,
					    malem_task_client_get_owner (task));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
malem_task_client_set_property (GObject      *object,
				guint         prop_id,
				const GValue *value,
				GParamSpec   *pspec)
{
	MalemTaskClient *task = MALEM_TASK_CLIENT (object);

	switch (prop_id)
	{
		case PROP_OBJECT_PATH:
			task->object_path = g_strdup (g_value_get_string (value));
			break;
		case PROP_STATUS:
			task->priv->status = g_value_get_enum (value);
			malem_task_status_changed (MALEM_TASK (object),
						   task->priv->status);
			break;
		case PROP_TITLE:
			task->priv->title = g_strdup (g_value_get_string (value));
			break;
		case PROP_PROGRESS:
			task->priv->progress = g_value_get_uint (value);
			task->priv->has_progress = TRUE;
			if (task->proxy != NULL)
				dbus_g_proxy_call_no_reply (task->proxy,
							    "SetProgress",
							    G_TYPE_UINT, task->priv->progress,
							    G_TYPE_INVALID);
			break;
		case PROP_OWNER:
			task->priv->owner = g_strdup(g_value_get_string (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
malem_task_client_start (MalemTask *task)
{
	MalemTaskClient *client = MALEM_TASK_CLIENT (task);
	if (client->proxy != NULL)
		dbus_g_proxy_call_no_reply (client->proxy,
					    "Start",
					    G_TYPE_INVALID);
}

static void
malem_task_client_pause (MalemTask *task)
{
	MalemTaskClient *client = MALEM_TASK_CLIENT (task);
	if (client->proxy != NULL)
		dbus_g_proxy_call_no_reply (client->proxy,
					    "Pause",
					    G_TYPE_INVALID);
}

static void
malem_task_client_abort (MalemTask *task)
{
	MalemTaskClient *client = MALEM_TASK_CLIENT (task);
	if (client->proxy != NULL)
		dbus_g_proxy_call_no_reply (client->proxy,
					    "Abort",
					    G_TYPE_INVALID);
}

static gboolean
malem_task_client_has_action (MalemTask   *task,
			      const gchar *action_name)
{
	MalemTaskClient *client = MALEM_TASK_CLIENT (task);
	GError *error = NULL;
	gboolean has_action;

	if (client->proxy == NULL)
		return FALSE;
	if (!dbus_g_proxy_call (client->proxy, "HasAction", &error,
				G_TYPE_STRING, action_name,
				G_TYPE_INVALID,
				G_TYPE_BOOLEAN, &has_action,
				G_TYPE_INVALID)) {
		g_warning ("Could not determine if the action '%s' exists for %s\n",
			   action_name,
			   client->object_path);
		return FALSE;
	}
	return has_action;
}

static void
malem_task_client_add_action (MalemTask   *task,
			      const gchar *action_name,
			      const gchar *label,
			      const gchar *icon_name)
{
	MalemTaskClient *client = MALEM_TASK_CLIENT (task);
	
	if (client->proxy != NULL) {
		if (label == NULL)
			label = "";
		if (icon_name == NULL)
			icon_name = "";
		dbus_g_proxy_call_no_reply (client->proxy,
					    "AddAction",
					    G_TYPE_STRING, action_name,
					    G_TYPE_STRING, label,
					    G_TYPE_STRING, icon_name,
					    G_TYPE_INVALID);
	}
}

static gchar **
malem_task_client_list_actions (MalemTask   *task)
{
	MalemTaskClient *client = MALEM_TASK_CLIENT (task);
	GError *error = NULL;
	gchar **interfaces;
	
	if (client->proxy == NULL)
		return NULL;

	if (!dbus_g_proxy_call (client->proxy, "ListActions", &error,
				G_TYPE_INVALID,
				G_TYPE_BOOLEAN, &interfaces,
				G_TYPE_INVALID)) {
		g_warning ("Could not determine get the action list for %s\n",
			   client->object_path);
		return NULL;
	}

	return interfaces;
}

static void
malem_task_client_trigger_action (MalemTask   *task,
				  const gchar *action_name)
{
	MalemTaskClient *client = MALEM_TASK_CLIENT (task);

	if (client->proxy != NULL)
		dbus_g_proxy_call_no_reply (client->proxy,
					    "TriggerAction",
					    G_TYPE_STRING, action_name,
					    G_TYPE_INVALID);
}

static void
malem_task_client_class_init (MalemTaskClientClass *klass)
{
	GObjectClass   *object_class = G_OBJECT_CLASS (klass);

	malem_task_client_parent_class = g_type_class_peek_parent (klass);

	object_class->constructor = malem_task_client_constructor;
	object_class->finalize = malem_task_client_finalize;
	object_class->get_property = malem_task_client_get_property;
	object_class->set_property = malem_task_client_set_property;

	g_object_class_override_property (object_class, PROP_STATUS, "status");
	g_object_class_override_property (object_class, PROP_TITLE, "title");
	g_object_class_override_property (object_class, PROP_PROGRESS, "progress");
	g_object_class_override_property (object_class, PROP_OWNER, "owner");

	g_object_class_install_property (object_class,
					 PROP_OBJECT_PATH,
					 g_param_spec_string ("object-path",
					 		      "Object Path",
					 		      "The task's object path",
					 		      "",
					 		      G_PARAM_READWRITE | 
					 		      G_PARAM_CONSTRUCT_ONLY));

	g_type_class_add_private (klass, sizeof (MalemTaskClientPrivate));
}

static void
malem_task_iface_init (MalemTaskIface *iface)
{
	iface->start = malem_task_client_start;
	iface->pause = malem_task_client_pause;
	iface->abort = malem_task_client_abort;
	iface->has_action = malem_task_client_has_action;
	iface->add_action = malem_task_client_add_action;
	iface->list_actions = malem_task_client_list_actions;
	iface->trigger_action = malem_task_client_trigger_action;
}

GType
malem_task_client_get_base_type (void)
{
	static GType the_type = 0;

	if (G_UNLIKELY (the_type == 0)) {
		static GTypeInfo info = {
			sizeof (MalemTaskClientClass),
			NULL, NULL,
			(GClassInitFunc) malem_task_client_class_init,
			NULL, NULL,
			sizeof (MalemTaskClient), 0,
			(GInstanceInitFunc) malem_task_client_init,
			NULL
		};

		static const GInterfaceInfo iface_info = {
			(GInterfaceInitFunc) malem_task_iface_init,
			NULL, NULL 
		};

		the_type = g_type_register_static (G_TYPE_INITIALLY_UNOWNED,
						   g_intern_static_string ("MalemTaskClient"),
						   &info,
						   (GTypeFlags) 0);

		g_type_add_interface_static (the_type,
					     MALEM_TYPE_TASK,
					     &iface_info);
	}
	return the_type;
}

static int
compare_quarks (const GQuark *q1,
		const GQuark *q2)
{
	return (*q1 > *q2) - (*q1 < *q2);
}

static gchar *
get_type_name (guint         n_ifaces, 
	       const GQuark *ifaces)
{
	guint i;
	GString *name;
	
	name = g_string_new ("MalemTaskClient");

	for (i = 0; i < n_ifaces; i++)
		g_string_append_printf (name, "+%u", ifaces[i]);
	
	return g_string_free (name, FALSE);
}

GType
malem_task_client_get_type (const gchar *ifaces[])
{
	static GHashTable *type_table = NULL;
	GType the_type = 0;
	guint n_ifaces, i;
	GQuark *ifaces_quarks;
	gchar *type_name;

	/* There is no additional interface: return base type */
	if (ifaces == NULL || *ifaces == NULL)
		return MALEM_TYPE_TASK_CLIENT;

	if (G_UNLIKELY (type_table == NULL))
		type_table = g_hash_table_new (g_str_hash, g_str_equal);

	n_ifaces = g_strv_length ((gchar**) ifaces);

	ifaces_quarks = g_new (GQuark, n_ifaces);
	for (i = 0; i < n_ifaces; i++)
		ifaces_quarks[i] = g_quark_from_string (ifaces[i]);
	qsort (ifaces_quarks, n_ifaces, sizeof (GQuark),
	       (GCompareFunc) compare_quarks);

	type_name = get_type_name (n_ifaces, ifaces_quarks);

	the_type = GPOINTER_TO_UINT (
			g_hash_table_lookup (type_table, (gpointer) type_name));
		
	if (G_UNLIKELY (the_type == 0)) {
		static GTypeInfo info = {
			sizeof (MalemTaskClientClass),
			NULL, NULL,
			NULL, NULL, NULL,
			sizeof (MalemTaskClient), 0,
			NULL, NULL
		};
		the_type = g_type_register_static (MALEM_TYPE_TASK_CLIENT,
						   type_name,
						   &info,
						   (GTypeFlags) 0);

		for (i = 0; i < n_ifaces; i++)
			malem_type_add_interface_static (the_type,
							 ifaces_quarks[i]);
		
		g_hash_table_insert (type_table,
				     (gpointer) g_strdup (type_name),
				     GUINT_TO_POINTER (the_type));
	}
	g_free (ifaces_quarks);
	g_free (type_name);
	return the_type;
}
