/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib/gi18n.h>
#include <gtk/gtk.h>

#include "malem-progress-window.h"
#include "malem-task-list-model.h"
#include "malem-task-list.h"

#define MALEM_PROGRESS_WINDOW_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o),	\
					      MALEM_TYPE_PROGRESS_WINDOW,	\
					      MalemProgressWindowPrivate))

struct _MalemProgressWindowPrivate {
	GtkWidget    *label;
	GtkWidget    *vbox;
	GtkWidget    *treeview;
	GtkWidget    *scrolled;
	GtkTreeModel *model;
	GtkWidget    *toolbar;
	GtkWidget    *pause_button;
	GtkWidget    *resume_button;
	GtkWidget    *abort_button;
};

enum {
	PROP_0,
	TASK_LIST
};

G_DEFINE_TYPE (MalemProgressWindow, malem_progress_window, MALEM_TYPE_ALIGNED_WINDOW)

static void
update_button_sensitivity (MalemProgressWindow *window,
			   GtkTreeIter         *iter)
{
	MalemTask *task;
	MalemTaskStatus status;

	gtk_tree_model_get (GTK_TREE_MODEL (window->priv->model), iter,
			    MALEM_TASK_LIST_MODEL_OBJECT, &task,
			    -1);
	
	status = malem_task_get_status (task);
	
	gtk_widget_set_sensitive (window->priv->resume_button,
				  (status == MALEM_TASK_STATUS_NEW ||
				   status == MALEM_TASK_STATUS_PAUSED) &&
				  malem_task_has_action (task, "start"));
	gtk_widget_set_sensitive (window->priv->pause_button,
				  status == MALEM_TASK_STATUS_STARTED &&
				  malem_task_has_action (task, "pause"));	
	gtk_widget_set_sensitive (window->priv->abort_button,
				  (status == MALEM_TASK_STATUS_NEW ||
				   status == MALEM_TASK_STATUS_STARTED ||
				   status == MALEM_TASK_STATUS_PAUSED) &&
				  malem_task_has_action (task, "abort"));
}

static void
on_selection_changed (GtkTreeSelection *selection,
		      MalemProgressWindow *window)
{
	GtkTreeModel *model;
	GList *selected_rows;
	GtkTreeIter iter;
	
	selected_rows = gtk_tree_selection_get_selected_rows (selection, &model);
	
	if (selected_rows == NULL)
		return; /* no selection */
	
	gtk_tree_model_get_iter (model, &iter, (GtkTreePath *) selected_rows->data);
	
	g_list_foreach (selected_rows, (GFunc) gtk_tree_path_free, NULL);
	g_list_free (selected_rows);
	
	update_button_sensitivity (window, &iter);
}

static void
status_func (GtkTreeViewColumn *column,
	     GtkCellRenderer   *cell,
	     GtkTreeModel      *model,
	     GtkTreeIter       *iter,
	     gpointer           user_data)
{
	MalemTaskStatus status;
	const gchar *icon_name = NULL;
	
	gtk_tree_model_get (model, iter,
			    MALEM_TASK_LIST_MODEL_STATUS, &status,
			    -1);

	switch (status)
	{
		case MALEM_TASK_STATUS_NEW:
			icon_name = GTK_STOCK_NEW;
			break;
		case MALEM_TASK_STATUS_STARTED:
			icon_name = GTK_STOCK_MEDIA_PLAY;
			break;
		case MALEM_TASK_STATUS_PAUSED:
			icon_name = GTK_STOCK_MEDIA_PAUSE;
			break;
		case MALEM_TASK_STATUS_ABORTED:
			icon_name = GTK_STOCK_STOP;
			break;
		case MALEM_TASK_STATUS_COMPLETED:
			icon_name = GTK_STOCK_APPLY;
			break;
	}
	
	g_object_set (cell, "stock-id", icon_name, NULL);
}

static GtkWidget *
create_tree_view (MalemProgressWindow *window)
{
	GtkTreeView *treeview;
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;
	GtkTreeSelection *selection;

	treeview = GTK_TREE_VIEW (gtk_tree_view_new ());

	renderer = gtk_cell_renderer_progress_new ();
	column = gtk_tree_view_column_new_with_attributes ("%", renderer,
							   "value", MALEM_TASK_LIST_MODEL_PROGRESS,
							   NULL);
	gtk_tree_view_append_column (treeview, column);
	renderer = gtk_cell_renderer_pixbuf_new ();

	column = gtk_tree_view_column_new_with_attributes ("S", renderer, NULL);
	gtk_tree_view_column_set_cell_data_func (column, renderer,
						 (GtkTreeCellDataFunc) status_func,
						 NULL, NULL);
	gtk_tree_view_append_column (treeview, column);

	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("Title", renderer,
							   "text", MALEM_TASK_LIST_MODEL_TITLE,
							   NULL);
	gtk_tree_view_append_column (treeview, column);

	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("Owner", renderer,
							   "text", MALEM_TASK_LIST_MODEL_OWNER,
							   NULL);
	gtk_tree_view_append_column (treeview, column);

	selection = gtk_tree_view_get_selection (treeview);
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_BROWSE);
	g_signal_connect (selection, "changed", 
			  G_CALLBACK (on_selection_changed),
			  window);
	
	return GTK_WIDGET (treeview);
}

static void
on_size_request (GtkWidget           *widget,
		 GtkRequisition      *requisition,
		 MalemProgressWindow *window)
{
	static gboolean recursion = FALSE;
	gint w, h, b;
	gint maxw, maxh;
	GdkScreen *screen;
	GtkRequisition req;

	if (recursion == TRUE)
		return;

	recursion = TRUE;

	screen = gtk_window_get_screen (GTK_WINDOW (window));

	maxw = 0.6 * gdk_screen_get_width (screen);
	maxh = 0.8 * gdk_screen_get_height (screen);

	gtk_widget_size_request (window->priv->treeview, &req);
	b = gtk_container_get_border_width (GTK_CONTAINER (window->priv->scrolled));
	
	/* take into account the border width and the shadow in the sizing */
	h = req.height + 2 + 2 * b;
	w = req.width + 2 + 2 * b;
	
	h = MIN (maxh, h);
	w = MIN (maxw, w);
	w = MAX (200, w);

	if (h > 0 && w > 0)
		gtk_widget_set_size_request (window->priv->scrolled, w, h);
	
	recursion = FALSE;
}

static void
update_view_visibility (MalemProgressWindow *window,
			gboolean             visible)
{
	if (visible) {
		if (GTK_WIDGET_VISIBLE (window->priv->label)) {
			gtk_widget_hide (window->priv->label);
			gtk_widget_show (window->priv->vbox);
		}
	} else {
		if (GTK_WIDGET_VISIBLE (window->priv->vbox)) {
			gtk_widget_hide (window->priv->vbox);
			gtk_widget_show (window->priv->label);
		}	
	}
}

static void
on_model_row_inserted (GtkTreeModel        *tree_model,
		       GtkTreePath         *path,
		       GtkTreeIter         *iter,
		       MalemProgressWindow *window)
{
	MalemTaskListModel *model = MALEM_TASK_LIST_MODEL (tree_model);
	gboolean visible = malem_task_list_get_length (model->tasklist) != 0;
	
	update_view_visibility (window, visible);
}

static void
on_model_row_deleted (GtkTreeModel        *tree_model,
		      GtkTreePath         *path,
		      MalemProgressWindow *window)
{
	MalemTaskListModel *model = MALEM_TASK_LIST_MODEL (tree_model);
	gboolean visible = malem_task_list_get_length (model->tasklist) != 1;
	
	update_view_visibility (window, visible);
}

static void
on_model_row_changed (GtkTreeModel        *tree_model,
		      GtkTreePath         *path,
		      GtkTreeIter         *iter,
		      MalemProgressWindow *window)
{
	GtkTreeSelection *selection;
	
	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (window->priv->treeview));
	if (gtk_tree_selection_get_selected (selection, NULL, iter))
		update_button_sensitivity (window, iter);
}

static void
set_task_list (MalemProgressWindow *window,
	       MalemTaskList       *tasklist)
{
	MalemTaskListModel *model;
	
	window->priv->model = malem_task_list_model_new (tasklist);

	gtk_tree_view_set_model (GTK_TREE_VIEW (window->priv->treeview),
				 window->priv->model);
	
	g_signal_connect_after (window->priv->model, "row-inserted",
				G_CALLBACK (on_model_row_inserted), window);
	g_signal_connect_after (window->priv->model, "row-deleted",
				G_CALLBACK (on_model_row_deleted), window);
	g_signal_connect_after (window->priv->model, "row-changed",
				G_CALLBACK (on_model_row_changed), window);

	model = MALEM_TASK_LIST_MODEL (window->priv->model);	
	update_view_visibility (window,
				malem_task_list_get_length (model->tasklist) != 0);
}

static void
on_action_button_clicked (GtkWidget           *button,
			  MalemProgressWindow *window)
{
	GtkTreeModel *model;
	GList *selected_rows;
	GtkTreeSelection *selection;
	GtkTreeIter iter;
	MalemTask *task;
	
	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (window->priv->treeview));
	selected_rows = gtk_tree_selection_get_selected_rows (selection, &model);
	
	if (selected_rows == NULL)
		return; /* no selection */
	
	gtk_tree_model_get_iter (model, &iter, (GtkTreePath *) selected_rows->data);
	
	g_list_foreach (selected_rows, (GFunc) gtk_tree_path_free, NULL);
	g_list_free (selected_rows);

	gtk_tree_model_get (model, &iter,
			    MALEM_TASK_LIST_MODEL_OBJECT, &task,
			    -1);

	if (button == window->priv->resume_button)
		malem_task_trigger_action (task, "start");
	else if (button == window->priv->pause_button)
		malem_task_trigger_action (task, "pause");
	else if (button == window->priv->abort_button)
		malem_task_trigger_action (task, "abort");
	else
		g_warning ("Unknown action button clicked\n");
}

static void
malem_progress_window_init (MalemProgressWindow *window)
{
	GtkWidget *box;
	GtkToolItem *tool_item;

	window->priv = MALEM_PROGRESS_WINDOW_GET_PRIVATE (window);

	/* This box contains two widgets that should never be shown at the
	   same time. */
	box = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (window), box);
	gtk_widget_show (box);
	
	/* Label (showed only when nothing in the treeview) */
	window->priv->label = gtk_label_new ("No active task");
	gtk_misc_set_padding (GTK_MISC (window->priv->label), 6, 2);
	gtk_container_add (GTK_CONTAINER (box), window->priv->label);
	gtk_widget_show (window->priv->label);
	
	/* TreeView window (showed otherwise) */
	window->priv->vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (box), window->priv->vbox);
	gtk_widget_show (window->priv->vbox);	

	/* TreeView */	
	window->priv->scrolled = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (window->priv->scrolled),
					     GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy  (GTK_SCROLLED_WINDOW (window->priv->scrolled),
                                         GTK_POLICY_NEVER,
                                         GTK_POLICY_AUTOMATIC);
	gtk_container_set_border_width (GTK_CONTAINER (window->priv->scrolled), 6);
	gtk_box_pack_start (GTK_BOX (window->priv->vbox), window->priv->scrolled,
			    TRUE, TRUE, 0);

	gtk_widget_show (window->priv->scrolled);

	g_signal_connect (window->priv->scrolled, "size-request",
	                  G_CALLBACK (on_size_request), window);

	window->priv->treeview = create_tree_view (window);
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (window->priv->treeview),
					   FALSE);
	gtk_container_add (GTK_CONTAINER (window->priv->scrolled),
			   window->priv->treeview);
	gtk_widget_show (window->priv->treeview);

	/* Toolbar */
	window->priv->toolbar = gtk_toolbar_new ();
	gtk_toolbar_set_style (GTK_TOOLBAR (window->priv->toolbar),
			       GTK_TOOLBAR_ICONS);
	gtk_toolbar_set_icon_size (GTK_TOOLBAR (window->priv->toolbar),
				   GTK_ICON_SIZE_MENU);
	gtk_box_pack_start (GTK_BOX (window->priv->vbox), window->priv->toolbar,
			    FALSE, TRUE, 0);
	gtk_widget_show (window->priv->toolbar);

	window->priv->resume_button = GTK_WIDGET (
			gtk_tool_button_new_from_stock (GTK_STOCK_MEDIA_PLAY));
	gtk_toolbar_insert (GTK_TOOLBAR (window->priv->toolbar),
			    GTK_TOOL_ITEM (window->priv->resume_button),
			    -1);
	gtk_widget_set_sensitive (window->priv->resume_button, FALSE);
	gtk_widget_show (window->priv->resume_button);
	g_signal_connect (window->priv->resume_button, "clicked",
			  G_CALLBACK (on_action_button_clicked), window);
	
	window->priv->pause_button = GTK_WIDGET (
			gtk_tool_button_new_from_stock (GTK_STOCK_MEDIA_PAUSE));
	gtk_toolbar_insert (GTK_TOOLBAR (window->priv->toolbar),
			    GTK_TOOL_ITEM (window->priv->pause_button),
			    -1);
	gtk_widget_set_sensitive (window->priv->pause_button, FALSE);
	gtk_widget_show (window->priv->pause_button);
	g_signal_connect (window->priv->pause_button, "clicked",
			  G_CALLBACK (on_action_button_clicked), window);

	window->priv->abort_button = GTK_WIDGET (
			gtk_tool_button_new_from_stock (GTK_STOCK_STOP));
	gtk_toolbar_insert (GTK_TOOLBAR (window->priv->toolbar),
			    GTK_TOOL_ITEM (window->priv->abort_button),
			    -1);
	gtk_widget_set_sensitive (window->priv->abort_button, FALSE);
	gtk_widget_show (window->priv->abort_button);
	g_signal_connect (window->priv->abort_button, "clicked",
			  G_CALLBACK (on_action_button_clicked), window);
	
	tool_item = gtk_separator_tool_item_new ();
	gtk_separator_tool_item_set_draw (GTK_SEPARATOR_TOOL_ITEM (tool_item), FALSE);
	gtk_tool_item_set_expand (tool_item, TRUE);
	gtk_toolbar_insert (GTK_TOOLBAR (window->priv->toolbar),
			    tool_item, -1);
	gtk_widget_show (GTK_WIDGET (tool_item));
	
	tool_item = gtk_tool_button_new_from_stock (GTK_STOCK_CLOSE);
	gtk_toolbar_insert (GTK_TOOLBAR (window->priv->toolbar),
			    tool_item, -1);
	gtk_widget_show (GTK_WIDGET (tool_item));
	g_signal_connect_swapped (tool_item, "clicked",
				  G_CALLBACK (gtk_widget_hide), window);
}

static void
malem_progress_window_set_property (GObject      *object,
				    guint         prop_id,
				    const GValue *value,
				    GParamSpec   *pspec)
{
	MalemProgressWindow *window = MALEM_PROGRESS_WINDOW (object);

	switch (prop_id)
	{
		case TASK_LIST:
			set_task_list (window, 
				       MALEM_TASK_LIST (g_value_get_object (value)));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
malem_progress_window_finalize (GObject *object)
{
	MalemProgressWindow *window = MALEM_PROGRESS_WINDOW (object);

	g_return_if_fail (window->priv != NULL);

	g_object_unref (window->priv->model);

	/* chainup */
	G_OBJECT_CLASS (malem_progress_window_parent_class)->finalize (object);
}

static void
malem_progress_window_class_init (MalemProgressWindowClass *klass)
{
	GObjectClass   *object_class = G_OBJECT_CLASS (klass);

	object_class->set_property = malem_progress_window_set_property;
	object_class->finalize	   = malem_progress_window_finalize;

	g_object_class_install_property (G_OBJECT_CLASS (klass), TASK_LIST,
					 g_param_spec_object ("task-list",
					 		      "Task List",
					 		      "The task list",
					 		      MALEM_TYPE_TASK_LIST,
					 		      G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));

	g_type_class_add_private (klass, sizeof (MalemProgressWindowPrivate));
}

GtkWidget *
malem_progress_window_new (MalemTaskList *tasklist,
			   GtkStatusIcon *icon)
{
	MalemProgressWindow *window;
	
	window = MALEM_PROGRESS_WINDOW (g_object_new (MALEM_TYPE_PROGRESS_WINDOW,
						      "status-icon", icon,
						      "task-list", tasklist,
						      NULL));
	
	return GTK_WIDGET (window);
}
