/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "malem-task-list-model.h"
#include "interfaces/malem-enum-types.h"

#define MALEM_TASK_LIST_MODEL_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o),	\
					      MALEM_TYPE_TASK_LIST_MODEL,	\
					      MalemTaskListModelPrivate))

struct _MalemTaskListModelPrivate {
	gulong         task_added_handler_id;
	gulong         task_removed_handler_id;
};

enum {
	PROP_0,
	PROP_TASK_LIST
};

static void		 on_task_changed			(MalemTask *task, guint data, MalemTaskListModel *model);
static GtkTreeModelFlags malem_task_list_model_get_flags	(GtkTreeModel *tree_model);
static gint		 malem_task_list_model_get_n_columns	(GtkTreeModel *tree_model);
static GType		 malem_task_list_model_get_column_type	(GtkTreeModel *tree_model, gint index);
static gboolean		 malem_task_list_model_get_iter		(GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreePath *path);
static GtkTreePath	*malem_task_list_model_get_path		(GtkTreeModel *tree_model, GtkTreeIter *iter);
static void		 malem_task_list_model_get_value	(GtkTreeModel *tree_model, GtkTreeIter *iter, gint column, GValue *value);
static gboolean		 malem_task_list_model_iter_next	(GtkTreeModel *tree_model, GtkTreeIter *iter);
static gboolean		 malem_task_list_model_iter_has_child	(GtkTreeModel *tree_model, GtkTreeIter *iter);
static gint		 malem_task_list_model_iter_n_children	(GtkTreeModel *tree_model, GtkTreeIter *iter);
static gboolean		 malem_task_list_model_iter_nth_child	(GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *parent, gint n);
static void		 gtk_tree_model_iface_init		(GtkTreeModelIface *iface);
static void		 malem_task_list_model_set_task_list	(MalemTaskListModel *model, MalemTaskList *tasklist);

G_DEFINE_TYPE_WITH_CODE(MalemTaskListModel, malem_task_list_model, G_TYPE_OBJECT,
			G_IMPLEMENT_INTERFACE(GTK_TYPE_TREE_MODEL,
					      gtk_tree_model_iface_init))

static void
malem_task_list_model_init (MalemTaskListModel *model)
{
	model->priv = MALEM_TASK_LIST_MODEL_GET_PRIVATE (model);
}

static void
malem_task_list_model_get_property (GObject    *object,
				    guint       prop_id,
				    GValue     *value,
				    GParamSpec *pspec)
{
	MalemTaskListModel *model = MALEM_TASK_LIST_MODEL (object);

	switch (prop_id)
	{
		case PROP_TASK_LIST:
			g_value_set_object (value, model->tasklist);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
malem_task_list_model_set_property (GObject      *object,
				    guint         prop_id,
				    const GValue *value,
				    GParamSpec   *pspec)
{
	MalemTaskListModel *model = MALEM_TASK_LIST_MODEL (object);

	switch (prop_id)
	{
		case PROP_TASK_LIST:
			malem_task_list_model_set_task_list (model, 
							     g_value_get_object (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
malem_task_list_model_finalize (GObject *object)
{
	MalemTaskListModel *model = MALEM_TASK_LIST_MODEL (object);
	
	g_signal_handler_disconnect (model->tasklist,
				     model->priv->task_added_handler_id);
	g_signal_handler_disconnect (model->tasklist,
				     model->priv->task_removed_handler_id);
	
	if (malem_task_list_get_length (model->tasklist) > 0) {
		MalemTaskListIter iter;
		
		malem_task_list_get_iter_first (model->tasklist, &iter);
		do {
			g_signal_handlers_disconnect_by_func (
				malem_task_list_iter_get_task (&iter),
				G_CALLBACK (on_task_changed),
				model);
		} while (malem_task_list_iter_next (&iter));		
	}
				     
	g_object_unref (model->tasklist);
	
	/* chainup */
	G_OBJECT_CLASS (malem_task_list_model_parent_class)->finalize (object);
}

static void
gtk_tree_model_iface_init (GtkTreeModelIface *iface)
{
	iface->get_flags = malem_task_list_model_get_flags;
	iface->get_n_columns = malem_task_list_model_get_n_columns;
	iface->get_column_type = malem_task_list_model_get_column_type;
	iface->get_iter = malem_task_list_model_get_iter;
	iface->get_path = malem_task_list_model_get_path;
	iface->get_value = malem_task_list_model_get_value;
	iface->iter_next = malem_task_list_model_iter_next;
	iface->iter_has_child = malem_task_list_model_iter_has_child;
	iface->iter_n_children = malem_task_list_model_iter_n_children;
	iface->iter_nth_child = malem_task_list_model_iter_nth_child;
}

static void
malem_task_list_model_class_init (MalemTaskListModelClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->get_property = malem_task_list_model_get_property;
	object_class->set_property = malem_task_list_model_set_property;
	object_class->finalize = malem_task_list_model_finalize;

	g_object_class_install_property (object_class,
					 PROP_TASK_LIST,
					 g_param_spec_object ("task-list",
							      "Task List",
							      "The task list",
							      MALEM_TYPE_TASK_LIST,
							      G_PARAM_READWRITE |
							      G_PARAM_CONSTRUCT_ONLY));

	g_type_class_add_private (klass, sizeof (MalemTaskListModelPrivate));
}

GtkTreeModel *
malem_task_list_model_new (MalemTaskList *tasklist)
{
	return GTK_TREE_MODEL (g_object_new (MALEM_TYPE_TASK_LIST_MODEL,
					     "task-list", tasklist,
					     NULL));
}

static void
set_tree_iter (GtkTreeIter       *tree_iter,
	       MalemTaskListIter *list_iter)
{
	tree_iter->stamp = list_iter->stamp;
	tree_iter->user_data = (gpointer) list_iter->llink;
	tree_iter->user_data2 = (gpointer) list_iter->tasklist;
}

static void
on_task_changed (MalemTask          *task,
		 guint               data,
		 MalemTaskListModel *model)
{
	MalemTaskListIter iter;
	GtkTreeIter  tree_iter;
	GtkTreePath *tree_path;
	
	malem_task_list_find (model->tasklist, &iter, task);
	set_tree_iter (&tree_iter, &iter);
	tree_path = malem_task_list_model_get_path (GTK_TREE_MODEL (model),
						    &tree_iter);
	gtk_tree_model_row_changed (GTK_TREE_MODEL (model),
				    tree_path, &tree_iter);
	gtk_tree_path_free (tree_path);
}

static void
connect_to_task_signals (MalemTaskListModel *model,
			 MalemTask *task)
{
	g_signal_connect_after (task, "progress-changed",
				G_CALLBACK (on_task_changed),
				model);
	g_signal_connect_after (task, "status-changed",
				G_CALLBACK (on_task_changed),
				model);
}

static void
on_task_added (MalemTaskList      *tasklist,
	       MalemTaskListIter  *iter,
	       MalemTaskListModel *model)
{
	GtkTreeIter  tree_iter;
	GtkTreePath *tree_path;
	MalemTask *task;

	set_tree_iter (&tree_iter, iter);
	tree_path = malem_task_list_model_get_path (GTK_TREE_MODEL (model),
						    &tree_iter);

	task = malem_task_list_iter_get_task (iter);
	connect_to_task_signals (model, task);

	gtk_tree_model_row_inserted (GTK_TREE_MODEL (model),
				     tree_path, &tree_iter);
	gtk_tree_path_free (tree_path);
}

static void
on_task_removed (MalemTaskList      *tasklist,
		 MalemTaskListIter  *iter,
		 MalemTaskListModel *model)
{
	GtkTreeIter  tree_iter;
	GtkTreePath *tree_path;
	MalemTask *task;

	task = malem_task_list_iter_get_task (iter);
	g_signal_handlers_disconnect_by_func (task,
					      G_CALLBACK (on_task_changed),
					      model);

	set_tree_iter (&tree_iter, iter);
	tree_path = malem_task_list_model_get_path (GTK_TREE_MODEL (model),
						    &tree_iter);

	gtk_tree_model_row_deleted (GTK_TREE_MODEL (model), tree_path);
	gtk_tree_path_free (tree_path);
}

static void
malem_task_list_model_set_task_list (MalemTaskListModel *model,
				     MalemTaskList      *tasklist)
{
	MalemTaskListIter iter;
	MalemTask *task;

	model->tasklist = tasklist;
	g_object_ref (tasklist);
	
	model->priv->task_added_handler_id =
		g_signal_connect (tasklist, "task-added",
				  G_CALLBACK (on_task_added),
				  model);

	model->priv->task_removed_handler_id = 
		g_signal_connect (tasklist, "task-removed",
				  G_CALLBACK (on_task_removed),
				  model);

	if (malem_task_list_get_length (tasklist) > 0) {
		malem_task_list_get_iter_first (tasklist, &iter);

		do {
			task = malem_task_list_iter_get_task (&iter);
			connect_to_task_signals (model, task);
		} while (malem_task_list_iter_next (&iter));
	}
}

static GtkTreeModelFlags
malem_task_list_model_get_flags (GtkTreeModel *tree_model)
{
	return GTK_TREE_MODEL_LIST_ONLY;
}

static gint
malem_task_list_model_get_n_columns (GtkTreeModel *tree_model)
{
	return MALEM_TASK_LIST_MODEL_N_COLUMNS;
}

static GType
malem_task_list_model_get_column_type (GtkTreeModel *tree_model,
				       gint          column)
{
	switch (column) 
	{
		case MALEM_TASK_LIST_MODEL_OBJECT:	return MALEM_TYPE_TASK;
		case MALEM_TASK_LIST_MODEL_TITLE:	return G_TYPE_STRING;
		case MALEM_TASK_LIST_MODEL_PROGRESS:	return G_TYPE_UINT;
		case MALEM_TASK_LIST_MODEL_STATUS:	return MALEM_TYPE_TASK_STATUS;
		case MALEM_TASK_LIST_MODEL_OWNER:	return G_TYPE_STRING;
		default:				return G_TYPE_INVALID;
	}
}

static gboolean
malem_task_list_model_get_iter (GtkTreeModel *tree_model,
				GtkTreeIter  *tree_iter,
				GtkTreePath  *path)
{
	MalemTaskListModel *model;
	MalemTaskListIter   iter;
	gint n;

	g_return_val_if_fail (MALEM_IS_TASK_LIST_MODEL (tree_model), FALSE);
//	g_return_val_if_fail (gtk_tree_path_get_depth (path) != 1, FALSE);

	model = MALEM_TASK_LIST_MODEL (tree_model);

	n = gtk_tree_path_get_indices (path)[0];

	if (G_UNLIKELY (n >= malem_task_list_get_length (model->tasklist)))
		return FALSE;

	malem_task_list_get_iter_at_position (model->tasklist, &iter, n);
	set_tree_iter (tree_iter, &iter);
	return TRUE;
}

static GtkTreePath *
malem_task_list_model_get_path (GtkTreeModel *tree_model,
				GtkTreeIter  *tree_iter)
{
	MalemTaskListModel *model;
	gint n = 0;

	g_return_val_if_fail (MALEM_IS_TASK_LIST_MODEL (tree_model), NULL);
	
	model = MALEM_TASK_LIST_MODEL (tree_model);

	g_return_val_if_fail (tree_iter != NULL, NULL);
	g_return_val_if_fail (tree_iter->stamp == model->tasklist->stamp, NULL);

	/* FIXME: Bad bad bad ! - take advandage of both types similarities */
	n = malem_task_list_iter_get_position ((MalemTaskListIter *) tree_iter);

	return gtk_tree_path_new_from_indices (n, -1);
}

static void
malem_task_list_model_get_value (GtkTreeModel *tree_model,
				 GtkTreeIter  *tree_iter,
				 gint          column,
				 GValue       *value)
{
	MalemTaskListModel *model;
	MalemTask *task;

	g_return_if_fail (MALEM_IS_TASK_LIST_MODEL (tree_model));
	
	model = MALEM_TASK_LIST_MODEL (tree_model);

	g_return_if_fail (tree_iter != NULL);
	g_return_if_fail (tree_iter->stamp == model->tasklist->stamp);

	/* FIXME: Bad bad bad ! - take advandage of both types similarities */
	task = malem_task_list_iter_get_task ((MalemTaskListIter *) tree_iter);
	
	g_assert (task != NULL);
	
	switch (column)
	{
		case MALEM_TASK_LIST_MODEL_OBJECT:
			g_value_init (value, MALEM_TYPE_TASK);
			g_value_set_object (value, task);
			break;
		case MALEM_TASK_LIST_MODEL_TITLE:
			g_value_init (value, G_TYPE_STRING);
			g_value_set_string (value, malem_task_get_title (task));
			break;
		case MALEM_TASK_LIST_MODEL_PROGRESS:
			g_value_init (value, G_TYPE_UINT);
			g_value_set_uint (value, malem_task_get_progress (task));
			break;
		case MALEM_TASK_LIST_MODEL_STATUS:
			g_value_init (value, MALEM_TYPE_TASK_STATUS);
			g_value_set_enum (value, malem_task_get_status (task));
			break;
		case MALEM_TASK_LIST_MODEL_OWNER:
			g_value_init (value, G_TYPE_STRING);
			g_value_set_string (value, malem_task_get_owner (task));
			break;
		default:
			break;
	}
}

static gboolean
malem_task_list_model_iter_next (GtkTreeModel *tree_model,
				 GtkTreeIter  *tree_iter)
{
	MalemTaskListModel *model;

	g_return_val_if_fail (MALEM_IS_TASK_LIST_MODEL (tree_model), FALSE);
	
	model = MALEM_TASK_LIST_MODEL (tree_model);

	g_return_val_if_fail (tree_iter != NULL, FALSE);
	g_return_val_if_fail (tree_iter->stamp == model->tasklist->stamp, FALSE);

	/* FIXME: Bad bad bad ! - take advandage of both types similarities */
	return malem_task_list_iter_next ((MalemTaskListIter *) tree_iter);
}

static gboolean
malem_task_list_model_iter_has_child (GtkTreeModel *tree_model,
				      GtkTreeIter  *tree_iter)
{
	return FALSE;
}

static gint
malem_task_list_model_iter_n_children (GtkTreeModel *tree_model,
				       GtkTreeIter  *tree_iter)
{
	g_return_val_if_fail (MALEM_IS_TASK_LIST_MODEL (tree_model), -1);

	if (G_LIKELY (tree_iter == NULL)) {
		MalemTaskListModel *model = MALEM_TASK_LIST_MODEL (tree_model);
		return malem_task_list_get_length (model->tasklist);
	}
	return -1;
}

static gboolean
malem_task_list_model_iter_nth_child (GtkTreeModel *tree_model,
				      GtkTreeIter  *tree_iter,
				      GtkTreeIter  *parent,
				      gint          n)
{
	MalemTaskListModel *model;
	MalemTaskListIter   iter;

	g_return_val_if_fail (MALEM_IS_TASK_LIST_MODEL (tree_model), FALSE);

	model = MALEM_TASK_LIST_MODEL (tree_model);

	if (G_UNLIKELY (parent != NULL))
		return FALSE;

	if (G_UNLIKELY (n >= malem_task_list_get_length (model->tasklist)))
		return FALSE;

	malem_task_list_get_iter_at_position (model->tasklist, &iter, n);
	set_tree_iter (tree_iter, &iter);
	return TRUE;
}
