/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib/gi18n.h>
#include <gtk/gtk.h>

#include "malem-tray-icon.h"
#include "malem-debug.h"

#define MALEM_TRAY_ICON_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o),	\
					MALEM_TYPE_TRAY_ICON,			\
					MalemTrayIconPrivate))

struct _MalemTrayIconPrivate
{
	GtkUIManager   *ui_manager;
	GtkActionGroup *actiongroup;
};

enum {
	QUIT,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0, };

static void menu_about_cb		(GtkAction *action, MalemTrayIcon *icon);
static void menu_quit_cb		(GtkAction *action, MalemTrayIcon *icon);

static GtkActionEntry menu_action_entries [] = {
	{ "TrayAbout", GTK_STOCK_ABOUT, N_("_About"), NULL,
	  NULL, G_CALLBACK (menu_about_cb) },
	{ "TrayQuit", GTK_STOCK_QUIT, N_("_Quit"), NULL,
	  NULL, G_CALLBACK (menu_quit_cb) }
};

G_DEFINE_TYPE (MalemTrayIcon, malem_tray_icon, GTK_TYPE_STATUS_ICON)

static void
menu_about_cb (GtkAction     *action,
	       MalemTrayIcon *icon)
{
	gtk_show_about_dialog (NULL,
			       "name", "Mathusalem",
			       "version", VERSION,
			       "copyright", "Copyright \xc2\xa9 2006 Steve Frécinaux",
			       "comments", "Long running task manager",
			       "logo-icon-name", "applications-other",
			       NULL);
}

static void menu_quit_cb (GtkAction     *action,
		          MalemTrayIcon *icon)
{
	malem_debug ("tray-icon", "Signal emitted: quit");
	g_signal_emit (icon, signals[QUIT], 0);
}

static void
malem_tray_icon_popup_menu (GtkStatusIcon *status_icon,
			    guint          button,
			    guint          activate_time)
{
	MalemTrayIcon *icon = MALEM_TRAY_ICON (status_icon);
	GtkWidget *popup;
	popup = gtk_ui_manager_get_widget (GTK_UI_MANAGER (icon->priv->ui_manager),
					   "/MalemTrayPopup");

	gtk_menu_popup (GTK_MENU (popup), NULL, NULL,
			gtk_status_icon_position_menu, status_icon,
			button, activate_time);
}

static GObject *
malem_tray_icon_constructor (GType                  type,
			     guint                  n_construct_properties,
			     GObjectConstructParam *construct_properties)
{
	MalemTrayIcon      *tray;
	MalemTrayIconClass *klass;
	GError		   *error = NULL;
	GtkWidget	   *widget;

	klass = MALEM_TRAY_ICON_CLASS (g_type_class_peek (MALEM_TYPE_TRAY_ICON));

	tray = MALEM_TRAY_ICON (G_OBJECT_CLASS (malem_tray_icon_parent_class)->constructor
			      (type, n_construct_properties,
			       construct_properties));

	tray->priv->actiongroup = gtk_action_group_new ("TrayActions");

	gtk_action_group_set_translation_domain (tray->priv->actiongroup,
						 GETTEXT_PACKAGE);

	gtk_action_group_add_actions (tray->priv->actiongroup,
				      menu_action_entries,
				      G_N_ELEMENTS (menu_action_entries),
				      tray);
	
	gtk_ui_manager_insert_action_group (tray->priv->ui_manager, tray->priv->actiongroup, 0);
	gtk_ui_manager_add_ui_from_string (tray->priv->ui_manager,
					   "<ui>"
					   "  <popup name=\"MalemTrayPopup\">"
					   "    <menuitem action=\"TrayAbout\" />"
					   "    <separator />"
					   "    <menuitem action=\"TrayQuit\" />"
					   "  </popup>"
					   "</ui>",
					   -1, &error);
	if (error != NULL) {
		g_warning ("Couldn't merge user interface for popup: %s", error->message);
		g_clear_error (&error);
	}

	gtk_ui_manager_ensure_update (tray->priv->ui_manager);

	/* Get notified of when the menu goes, as we have to re-enable the tooltip */
	widget = gtk_ui_manager_get_widget (tray->priv->ui_manager, "/MalemTrayPopup");

	g_object_unref (tray->priv->actiongroup);

	return G_OBJECT (tray);
}

static void
malem_tray_icon_init (MalemTrayIcon *icon)
{
	icon->priv = MALEM_TRAY_ICON_GET_PRIVATE (icon);

	icon->priv->ui_manager = gtk_ui_manager_new ();


	gtk_status_icon_set_from_icon_name (GTK_STATUS_ICON (icon),
					    "applications-other");

	malem_debug ("tray-icon", "MalemTrayIcon Initialized");
}

static void
malem_tray_icon_finalize (GObject *object)
{
	MalemTrayIcon *icon = MALEM_TRAY_ICON (object);

	g_return_if_fail (icon->priv != NULL);

	malem_debug ("tray-icon", "MalemTrayIcon Finalized");

	/* chainup */
	G_OBJECT_CLASS (malem_tray_icon_parent_class)->finalize (object);
}

static void
malem_tray_icon_class_init (MalemTrayIconClass *klass)
{
	GType the_type = G_TYPE_FROM_CLASS (klass);
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	GtkStatusIconClass *icon_class = GTK_STATUS_ICON_CLASS (klass);

	object_class->finalize	   = malem_tray_icon_finalize;
	object_class->constructor  = malem_tray_icon_constructor;

	icon_class->popup_menu = malem_tray_icon_popup_menu;

	signals[QUIT] =
		g_signal_new ("quit",
			      the_type,
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (MalemTrayIconClass, quit_menu),
			      NULL,
			      NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);

	g_type_class_add_private (klass, sizeof (MalemTrayIconPrivate));
}

MalemTrayIcon *
malem_tray_icon_new (void)
{
	MalemTrayIcon *icon;
	icon = g_object_new (MALEM_TYPE_TRAY_ICON, NULL);
	return MALEM_TRAY_ICON (icon);
}

