/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <glib.h>
#include <dbus/dbus-glib.h>
#include "malem-dictionary.h"
#include "client/malem-task-client.h"

#define PROXY_DATA_KEY      "malem-dictionary-dbus-proxy"
#define HASH_TABLE_DATA_KEY "malem-dictionary-hash-table"

static GValue *
value_copy (const GValue *src_value)
{
	GValue *dest_value = g_new0 (GValue, 1);
	
	if (G_VALUE_TYPE (src_value)) {
		g_value_init (dest_value, G_VALUE_TYPE (src_value));
		g_value_copy (src_value, dest_value);
	}
	
	return dest_value;
}

static void
value_destroy (GValue *value)
{
	g_value_unset (value);
	g_free (value);
}

static GHashTable *
get_hash_table (MalemDictionary *dictionary)
{
	GHashTable *table;
	
	table = (GHashTable *) g_object_get_data (G_OBJECT (dictionary),
						  HASH_TABLE_DATA_KEY);

	if (table == NULL) {
		table = g_hash_table_new_full (g_str_hash, g_str_equal,
					       (GDestroyNotify) g_free,
					       (GDestroyNotify) value_destroy);
		g_object_set_data_full (G_OBJECT (dictionary),
					HASH_TABLE_DATA_KEY,
					(gpointer) table,
					(GDestroyNotify) g_hash_table_unref);
	}

	return table;
}

static void
on_dbus_value_added (DBusGProxy      *proxy,
		     gchar           *key,
		     MalemDictionary *dictionary)
{
	malem_dictionary_value_added (dictionary, key);
}

static void
on_dbus_value_changed (DBusGProxy      *proxy,
		       gchar           *key,
		       MalemDictionary *dictionary)
{
	g_hash_table_remove (get_hash_table (dictionary), key);
	malem_dictionary_value_changed (dictionary, key);
}

static void
on_dbus_key_removed (DBusGProxy      *proxy,
		     gchar           *key,
		     MalemDictionary *dictionary)
{
	g_hash_table_remove (get_hash_table (dictionary), key);
	malem_dictionary_key_removed (dictionary, key);
}

static DBusGProxy *
get_proxy (MalemDictionary *dictionary)
{
	DBusGProxy *proxy;
	MalemTaskClient *client = MALEM_TASK_CLIENT (dictionary);

	/* return null if task terminated */
	if (client->proxy == NULL)
		return NULL;

	proxy = DBUS_G_PROXY (g_object_get_data (G_OBJECT (client),
						 PROXY_DATA_KEY));

	if (proxy == NULL) {
		DBusGConnection *connection;
	        GError *error = NULL;
	
		connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);

		if (connection == NULL) {
			g_printerr ("Failed to open connection to bus: %s\n",
        	        	    error->message);
			g_error_free (error);
			return NULL;
		}
	
		proxy = dbus_g_proxy_new_for_name (connection,
						   "org.gnome.Mathusalem",
						   client->object_path,
						   "org.gnome.Mathusalem.Dictionary");

		dbus_g_proxy_add_signal (proxy,
					 "ValueAdded",
					 G_TYPE_STRING,
					 G_TYPE_INVALID);
	
		dbus_g_proxy_connect_signal (proxy,
					     "ValueAdded",
					     G_CALLBACK (on_dbus_value_added),
					     client, NULL);

		dbus_g_proxy_add_signal (proxy,
					 "ValueChanged",
					 G_TYPE_STRING,
					 G_TYPE_INVALID);
	
		dbus_g_proxy_connect_signal (proxy,
					     "ValueChanged",
					     G_CALLBACK (on_dbus_value_changed),
					     client, NULL);

		dbus_g_proxy_add_signal (proxy,
					 "KeyRemoved",
					 G_TYPE_STRING,
					 G_TYPE_INVALID);
	
		dbus_g_proxy_connect_signal (proxy,
					     "KeyRemoved",
					     G_CALLBACK (on_dbus_key_removed),
					     client, NULL);

		g_object_set_data_full (G_OBJECT (client),
					PROXY_DATA_KEY,
					(gpointer) proxy,
					(GDestroyNotify) g_object_unref);
	}

	return proxy;
}

static gboolean
impl_get_value (MalemDictionary *dictionary,
		const gchar     *key,
		GValue          *value)
{
	GValue *stored_value;
	GHashTable *table;
	
	table = get_hash_table (dictionary);
	stored_value = g_hash_table_lookup (table, key);

	if (stored_value == NULL) {
		DBusGProxy *proxy = get_proxy (dictionary);
		GError *error = NULL;
		
		if (proxy == NULL)
			return FALSE;
		
		if (!dbus_g_proxy_call (proxy, "GetValue", &error,
					G_TYPE_STRING, key,
					G_TYPE_INVALID,
					G_TYPE_VALUE, &stored_value,
					G_TYPE_INVALID)) {
			g_printerr ("Error while getting '%s' value: %s\n", 
				    key,
				    error->message);
			g_error_free (error);
			return FALSE;			
		}
		g_hash_table_replace (table,
				      g_strdup (key),
				      value_copy (stored_value));
	}
	
	g_value_init (value, G_VALUE_TYPE (stored_value));
	g_value_copy (stored_value, value);
	return TRUE;
}

static void
impl_set_value (MalemDictionary *dictionary,
		const gchar     *key,
		const GValue    *value)
{
	DBusGProxy *proxy = get_proxy (dictionary);
	
	g_return_if_fail (proxy != NULL);

	dbus_g_proxy_call_no_reply (proxy, "SetValue",
				    G_TYPE_STRING, key,
				    G_TYPE_VALUE, value,
				    G_TYPE_INVALID);
}

static void
impl_remove_key (MalemDictionary *dictionary,
		 const gchar     *key)
{
	DBusGProxy *proxy = get_proxy (dictionary);
	
	g_return_if_fail (proxy != NULL);

	dbus_g_proxy_call_no_reply (proxy, "SetValue",
				    G_TYPE_STRING, key,
				    G_TYPE_INVALID);
}

static gboolean
impl_has_key (MalemDictionary *dictionary,
	      const gchar     *key)
{
	if (g_hash_table_lookup (get_hash_table (dictionary), key) == NULL) {
		/* the key is stored locally -> it obviously exists */
		return FALSE;
	} else {
		/* check on the server */
		DBusGProxy *proxy = get_proxy (dictionary);
		GError *error = NULL;
		gboolean exists;
		
		if (proxy == NULL)
			return FALSE;

		if (!dbus_g_proxy_call (proxy, "HasKey", &error,
					G_TYPE_STRING, key,
					G_TYPE_INVALID,
					G_TYPE_BOOLEAN, &exists,
					G_TYPE_INVALID)) {
			g_printerr ("Error while checking '%s' existence: %s\n", 
				    key,
				    error->message);
			g_error_free (error);
			return FALSE;			
		}
		return exists;
	}
}

static gchar **
impl_list_keys (MalemDictionary *dictionary)
{
	DBusGProxy *proxy = get_proxy (dictionary);
	gchar **keys;
	GError *error = NULL;
	
	if (!dbus_g_proxy_call (proxy, "ListKeys", &error,
				G_TYPE_INVALID,
				G_TYPE_STRV, &keys,
				G_TYPE_INVALID)) {
		g_printerr ("Error while retrieving key list: %s\n", 
			    error->message);
		g_error_free (error);
		return NULL;
	}
	
	return keys;
}

void
malem_dictionary_init_default (MalemDictionary *dictionary)
{
	get_proxy (dictionary);
}


void
malem_dictionary_iface_init_default (MalemDictionaryIface *iface)
{
	iface->get_value = impl_get_value;
	iface->set_value = impl_set_value;
	iface->remove_key = impl_remove_key;
	iface->has_key = impl_has_key;
	iface->list_keys = impl_list_keys;
}
