/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <glib.h>
#include "malem-dictionary.h"

#define HASH_TABLE_DATA_KEY "malem-dictionary-hash-table"

static GValue *
value_copy (const GValue *src_value)
{
	GValue *dest_value = g_new0 (GValue, 1);
	
	if (G_VALUE_TYPE (src_value)) {
		g_value_init (dest_value, G_VALUE_TYPE (src_value));
		g_value_copy (src_value, dest_value);
	}
	
	return dest_value;
}

static void
value_destroy (GValue *value)
{
	g_value_unset (value);
	g_free (value);
}

static GHashTable *
get_hash_table (MalemDictionary *dictionary)
{
	GHashTable *table;
	
	table = (GHashTable *) g_object_get_data (G_OBJECT (dictionary),
						  HASH_TABLE_DATA_KEY);

	if (table == NULL) {
		table = g_hash_table_new_full (g_str_hash, g_str_equal,
					       (GDestroyNotify) g_free,
					       (GDestroyNotify) value_destroy);
		g_object_set_data_full (G_OBJECT (dictionary),
					HASH_TABLE_DATA_KEY,
					(gpointer) table,
					(GDestroyNotify) g_hash_table_unref);
	}

	return table;
}

static gboolean
get_value (MalemDictionary *dictionary,
	   const gchar     *key,
	   GValue          *value)
{
	GValue *stored_value;
	
	stored_value = g_hash_table_lookup (get_hash_table (dictionary), key);

	if (stored_value != NULL) {
		g_value_init (value, G_VALUE_TYPE (stored_value));
		g_value_copy (stored_value, value);
		return TRUE;
	} else {
		return FALSE;
	}
}

static void
set_value (MalemDictionary *dictionary,
	   const gchar     *key,
	   const GValue    *value)
{
	GHashTable *table;

	table = get_hash_table (dictionary);

	if (g_hash_table_lookup (table, key)) {
		g_hash_table_replace (table, g_strdup (key), value_copy (value));
		malem_dictionary_value_changed (dictionary, key);
	} else {
		g_hash_table_insert (table, g_strdup (key), value_copy (value));
		malem_dictionary_value_added (dictionary, key);
	}
}

static void
remove_key (MalemDictionary *dictionary,
	    const gchar     *key)
{
	if (g_hash_table_remove (get_hash_table (dictionary), key))
		malem_dictionary_key_removed (dictionary, key);
}

static gboolean
has_key (MalemDictionary *dictionary,
	 const gchar     *key)
{
	return g_hash_table_lookup (get_hash_table (dictionary), key) != NULL;
}

static void
get_key_foreach_func (const gchar *key,
		      const GValue *value,
		      GArray       *array)
{
	gchar *key_dup = g_strdup (key);
	g_array_append_val (array, key_dup);
}

static gchar **
list_keys (MalemDictionary *dictionary)
{
	GArray *array;
	
	array = g_array_new (TRUE, FALSE, sizeof (gchar *));

	g_hash_table_foreach (get_hash_table (dictionary),
			      (GHFunc) get_key_foreach_func,
			      array);
	g_array_sort (array, (GCompareFunc) g_ascii_strcasecmp);

	return (gchar **) g_array_free (array, FALSE);
}

void
malem_dictionary_init_default (MalemDictionary *dictionary)
{
}

void
malem_dictionary_iface_init_default (MalemDictionaryIface *iface)
{
	iface->get_value = get_value;
	iface->set_value = set_value;
	iface->remove_key = remove_key;
	iface->has_key = has_key;
	iface->list_keys = list_keys;
}
