/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif
#include "malem-dictionary.h"
#include "malem-task.h"

enum {
	VALUE_ADDED,
	VALUE_CHANGED,
	KEY_REMOVED,

	LAST_SIGNAL
};

static guint dictionary_signals[LAST_SIGNAL];

enum {
	ERROR_NON_EXISTENT_KEY = 1
};

static gboolean malem_dictionary_dbus_get_value		(MalemDictionary *dictionary, const gchar *key, GValue *value,	GError **error);
static gboolean malem_dictionary_dbus_set_value		(MalemDictionary *dictionary, const gchar *key, const GValue *value, GError **error);
static gboolean malem_dictionary_dbus_remove_key	(MalemDictionary *dictionary, const gchar *key, GError **error);
static gboolean malem_dictionary_dbus_has_key		(MalemDictionary *dictionary, const gchar *key, gboolean *has_key, GError **error);
static gboolean malem_dictionary_dbus_list_keys		(MalemDictionary *dictionary, gchar ***keys, GError **error);
#include "malem-dictionary-dbus-glue.h"

static void
malem_dictionary_class_init (gpointer g_iface)
{
	GType iface_type = G_TYPE_FROM_INTERFACE (g_iface);
	
	dictionary_signals[VALUE_ADDED] =
		g_signal_new ("value-added",
			      iface_type,
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (MalemDictionaryIface, value_added),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE,
			      1, G_TYPE_STRING);

	dictionary_signals[VALUE_CHANGED] =
		g_signal_new ("value-changed",
			      iface_type,
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (MalemDictionaryIface, value_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE,
			      1, G_TYPE_STRING);

	dictionary_signals[KEY_REMOVED] =
		g_signal_new ("key-removed",
			      iface_type,
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (MalemDictionaryIface, key_removed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE,
			      1, G_TYPE_STRING);

	dbus_g_object_type_install_info (iface_type,
					 &dbus_glib_malem_dictionary_object_info);
}

GType
malem_dictionary_get_type (void)
{
	static GType the_type = 0;
	
	if (G_UNLIKELY (the_type == 0)) {
		static const GTypeInfo info = {
			sizeof (MalemDictionaryIface),
			NULL, NULL,
			(GClassInitFunc) malem_dictionary_class_init,
			NULL, NULL, 0, 0, NULL
		};
		
		the_type = g_type_register_static (G_TYPE_INTERFACE,
						   g_intern_static_string ("MalemDictionary"),
						   &info, 0);
		g_type_interface_add_prerequisite (the_type, MALEM_TYPE_TASK);
	}
	return the_type;
}

gboolean
malem_dictionary_get_value (MalemDictionary *dictionary,
			    const gchar     *key,
			    GValue          *value)
{
	g_return_val_if_fail (MALEM_IS_DICTIONARY (dictionary), FALSE);

	return (* MALEM_DICTIONARY_GET_IFACE (dictionary)->get_value) (dictionary, key, value);
}

void
malem_dictionary_set_value (MalemDictionary *dictionary,
			    const gchar     *key,
			    const GValue    *value)
{
	g_return_if_fail (MALEM_IS_DICTIONARY (dictionary));

	(* MALEM_DICTIONARY_GET_IFACE (dictionary)->set_value) (dictionary, key, value);
}

void
malem_dictionary_remove_key (MalemDictionary *dictionary,
			     const gchar     *key)
{
	g_return_if_fail (MALEM_IS_DICTIONARY (dictionary));

	(* MALEM_DICTIONARY_GET_IFACE (dictionary)->remove_key) (dictionary, key);
}

gboolean
malem_dictionary_has_key (MalemDictionary *dictionary,
			  const gchar     *key)
{
	g_return_val_if_fail (MALEM_IS_DICTIONARY (dictionary), FALSE);

	return (* MALEM_DICTIONARY_GET_IFACE (dictionary)->has_key) (dictionary, key);
}

gchar **
malem_dictionary_list_keys (MalemDictionary *dictionary)
{
	g_return_val_if_fail (MALEM_IS_DICTIONARY (dictionary), NULL);

	return (* MALEM_DICTIONARY_GET_IFACE (dictionary)->list_keys) (dictionary);
}

void
malem_dictionary_value_added (MalemDictionary *dictionary,
			      const gchar     *key)
{
	g_return_if_fail (MALEM_IS_DICTIONARY (dictionary));

	g_signal_emit (dictionary, dictionary_signals[VALUE_ADDED], 0, key);
}

void
malem_dictionary_value_changed (MalemDictionary *dictionary,
			        const gchar     *key)
{
	g_return_if_fail (MALEM_IS_DICTIONARY (dictionary));

	g_signal_emit (dictionary, dictionary_signals[VALUE_CHANGED], 0, key);
}

void
malem_dictionary_key_removed (MalemDictionary *dictionary,
			      const gchar     *key)
{
	g_return_if_fail (MALEM_IS_DICTIONARY (dictionary));

	g_signal_emit (dictionary, dictionary_signals[KEY_REMOVED], 0, key);
}

/* D-Bus */

static GQuark
malem_dictionary_error_domain_quark (void)
{
	static GQuark quark = 0;
	
	if (!quark)
		quark = g_quark_from_static_string ("MalemDictionnaryError");
	return quark;
}

static gboolean
malem_dictionary_dbus_get_value (MalemDictionary  *dictionary,
				 const gchar      *key,
				 GValue           *value,
				 GError          **error)
{
	if (malem_dictionary_get_value (dictionary, key, value)) {
		return TRUE;
	} else {
		g_set_error (error, malem_dictionary_error_domain_quark (),
			     ERROR_NON_EXISTENT_KEY,
			     "Key \"%s\" does not exist", key);
		return FALSE;
	}
}

static gboolean
malem_dictionary_dbus_set_value (MalemDictionary  *dictionary,
				 const gchar      *key,
				 const GValue     *value,
				 GError          **error)
{
	malem_dictionary_set_value (dictionary, key, value);
	return TRUE;
}

static gboolean
malem_dictionary_dbus_remove_key (MalemDictionary  *dictionary,
				  const gchar      *key,
				  GError          **error)
{
	if (!malem_dictionary_has_key (dictionary, key)) {
		g_set_error (error, malem_dictionary_error_domain_quark (),
			     ERROR_NON_EXISTENT_KEY,
			     "Key \"%s\" does not exist", key);
		return FALSE;
	}

	malem_dictionary_remove_key (dictionary, key);
	return TRUE;
}

static gboolean
malem_dictionary_dbus_has_key (MalemDictionary  *dictionary,
			       const gchar      *key,
			       gboolean         *has_key,
			       GError          **error)
{
	*has_key = malem_dictionary_has_key (dictionary, key);
	return TRUE;
}

static gboolean
malem_dictionary_dbus_list_keys (MalemDictionary   *dictionary,
				 gchar           ***keys,
				 GError           **error)
{
	*keys = malem_dictionary_list_keys (dictionary);
	return TRUE;
}
