/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif
#include <dbus/dbus-glib.h>
#include <glib-object.h>
#include "malem-factory.h"
#include "malem-marshal.h"

enum {
	TASK_REGISTERED,

	LAST_SIGNAL
};

static guint factory_signals[LAST_SIGNAL];

static void malem_factory_class_init (gpointer g_iface);

static gboolean malem_factory_dbus_register_task	(MalemFactory *factory, const gchar **interfaces, const gchar *title, 
							 const gchar *owner, gchar **task_path, GError **error);
static gboolean malem_factory_dbus_list_tasks		(MalemFactory *factory, gchar ***task_paths, GError **error);
static gboolean malem_factory_dbus_list_tasks_for_interface
							(MalemFactory *factory, const gchar *interface, gchar ***task_paths,
							 GError **error);
static gboolean	malem_factory_dbus_get_version		(MalemFactory *factory,	GArray **version, GError **error);
#include "malem-factory-dbus-glue.h"

GType
malem_factory_get_type (void)
{
	static GType the_type = 0;
	
	if (G_UNLIKELY (the_type == 0)) {
		static GTypeInfo type_info = {
			sizeof (MalemFactoryIface),
			NULL, NULL,
			(GClassInitFunc) malem_factory_class_init
		};
		the_type = g_type_register_static (G_TYPE_INTERFACE,
						   g_intern_static_string ("MalemFactory"),
						   &type_info, 0);
		g_type_interface_add_prerequisite (the_type, G_TYPE_OBJECT);
	}
	return the_type;
}

static void
malem_factory_class_init (gpointer g_iface)		
{
	GType iface_type = G_TYPE_FROM_INTERFACE (g_iface);

	factory_signals[TASK_REGISTERED] =
		g_signal_new ("task-registered",
			      iface_type,
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (MalemFactoryIface, task_registered),
			      NULL, NULL,
			      malem_marshal_VOID__OBJECT_POINTER,
			      G_TYPE_NONE,
			      2, G_TYPE_OBJECT, G_TYPE_STRV);

	dbus_g_object_type_install_info (iface_type,
					 &dbus_glib_malem_factory_object_info);
}

MalemTask *
malem_factory_register_task (MalemFactory  *factory,
			     const gchar  **interfaces,
			     const gchar   *title,
			     const gchar   *owner)
{
	g_return_val_if_fail (MALEM_IS_FACTORY (factory), NULL);

	return (* MALEM_FACTORY_GET_IFACE (factory)->register_task) (factory, interfaces, 
								     title, owner);
}

MalemTask **
malem_factory_list_tasks (MalemFactory *factory)
{
	g_return_val_if_fail (MALEM_IS_FACTORY (factory), NULL);

	return (* MALEM_FACTORY_GET_IFACE (factory)->list_tasks) (factory);
}

MalemTask **
malem_factory_list_tasks_for_interface (MalemFactory *factory,
					const gchar  *interface)
{
	g_return_val_if_fail (MALEM_IS_FACTORY (factory), NULL);

	return (* MALEM_FACTORY_GET_IFACE (factory)->list_tasks_for_interface)
							(factory, interface);
}

void
malem_factory_get_version (MalemFactory  *factory,
			   gint          *major,
			   gint          *minor,
			   gint          *micro)
{
	g_return_if_fail (MALEM_IS_FACTORY (factory));

	(* MALEM_FACTORY_GET_IFACE (factory)->get_version) (factory, major, minor, micro);
}

void
malem_factory_task_registered (MalemFactory     *factory,
			       MalemTask        *task,
			       const gchar     **interfaces)
{
	g_return_if_fail (MALEM_IS_FACTORY (factory));
	g_return_if_fail (MALEM_IS_TASK (task));

	g_signal_emit (factory, factory_signals[TASK_REGISTERED], 0, 
		       G_OBJECT (task), interfaces);
}

/* DBus */

static gboolean
malem_factory_dbus_register_task (MalemFactory *factory,
				  const gchar **interfaces,
				  const gchar *title, 
				  const gchar *owner,
				  gchar **task_path,
				  GError **error)
{
	MalemTask *task;
	
	task = malem_factory_register_task (factory, interfaces, title, owner);
	*task_path = g_strdup (g_object_get_data (G_OBJECT (task),
			       "dbus_glib_object_path"));
	return TRUE;
}

static gchar **
get_paths_from_tasks (MalemTask **tasks)
{
	GPtrArray *paths;
	gchar *task_path;
	gint i;
	
	paths = g_ptr_array_new();
	for (i = 0; tasks[i] != NULL; i++) {
		task_path = g_object_get_data (G_OBJECT (tasks[i]),
					       "dbus_glib_object_path");
		g_ptr_array_add (paths, g_strdup (task_path));
	}
	/* NULL-terminate the array */
	g_ptr_array_add (paths, NULL);
	return (gchar **) g_ptr_array_free (paths, FALSE);
}

static gboolean
malem_factory_dbus_list_tasks (MalemFactory   *factory,
			       gchar        ***task_paths,
			       GError        **error)
{
	MalemTask **tasks;
	tasks = malem_factory_list_tasks (factory);
	*task_paths =  get_paths_from_tasks (tasks);
	g_free (tasks);
	return TRUE;}

static gboolean
malem_factory_dbus_list_tasks_for_interface (MalemFactory   *factory,
					     const gchar    *interface,
					     gchar        ***task_paths,
					     GError        **error)
{
	MalemTask **tasks;
	tasks = malem_factory_list_tasks_for_interface (factory, interface);
	*task_paths =  get_paths_from_tasks (tasks);
	g_free (tasks);
	return TRUE;
}

static gboolean
malem_factory_dbus_get_version (MalemFactory  *factory,
				GArray       **version,
				GError       **error)
{
	gint *v;
	*version = g_array_sized_new (FALSE, FALSE, sizeof(gint), 3);
	g_array_set_size (*version, 3);
	v = (gint *) (*version)->data;	
	malem_factory_get_version (factory, v, v + 1, v + 2);
	return TRUE;
}
