/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <glib.h>
#include <dbus/dbus-glib.h>
#include "malem-file-transfer.h"
#include "client/malem-task-client.h"

#define PROXY_DATA_KEY "malem-file-transfer-dbus-proxy"

static void
on_dbus_file_added (DBusGProxy        *proxy,
		    gchar             *filename,
		    MalemFileTransfer *transfer)
{
	malem_file_transfer_file_added (transfer, filename);
}

static DBusGProxy *
get_proxy (MalemFileTransfer *transfer)
{
	DBusGProxy *proxy;
	MalemTaskClient *client = MALEM_TASK_CLIENT (transfer);

	/* return null if task terminated */
	if (client->proxy == NULL) {
		return NULL;
	}

	proxy = DBUS_G_PROXY (g_object_get_data (G_OBJECT (client),
						 PROXY_DATA_KEY));

	if (proxy == NULL) {
		DBusGConnection *connection;
	        GError *error = NULL;
	
		connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);

		if (connection == NULL) {
			g_printerr ("Failed to open connection to bus: %s\n",
        	        	    error->message);
			g_error_free (error);
			return NULL;
		}

		proxy = dbus_g_proxy_new_for_name (connection,
						   "org.gnome.Mathusalem",
						   client->object_path,
						   "org.gnome.Mathusalem.FileTransfer");

		dbus_g_proxy_add_signal (proxy,
					 "FileAdded",
					 G_TYPE_STRING,
					 G_TYPE_INVALID);

		dbus_g_proxy_connect_signal (proxy,
					     "FileAdded",
					     G_CALLBACK (on_dbus_file_added),
					     client, NULL);

		g_object_set_data_full (G_OBJECT (client),
					PROXY_DATA_KEY,
					(gpointer) proxy,
					(GDestroyNotify) g_object_unref);
	}

	return proxy;
}

static void
impl_add_file (MalemFileTransfer          *transfer,
	       const gchar                *filename,
	       MalemFileTransferDirection  direction)
{
	DBusGProxy *proxy;
	GEnumClass *klass;
	GEnumValue *evalue;
	
	proxy = get_proxy (transfer);
	
	if (proxy != NULL) {
		klass = g_type_class_ref (MALEM_TYPE_FILE_TRANSFER_DIRECTION);
		evalue = g_enum_get_value (klass, direction);
		dbus_g_proxy_call_no_reply (proxy, "AddFile",
					    G_TYPE_STRING, filename,
					    G_TYPE_STRING, evalue->value_nick,
					    G_TYPE_INVALID);
		g_type_class_unref (klass);
	}
}

static MalemFileTransferDirection
impl_get_file_direction (MalemFileTransfer  *transfer,
			 const gchar        *filename)
{
	DBusGProxy *proxy;
	GEnumClass *klass;
	GEnumValue *evalue;
	GError *error = NULL;
	gchar *dir;
	MalemFileTransferDirection direction;
	
	proxy = get_proxy (transfer);
	
	if (proxy != NULL) {	
		if (!dbus_g_proxy_call (proxy, "GetFileDirection", &error,
					G_TYPE_STRING, filename,
					G_TYPE_INVALID,
					G_TYPE_STRING, &dir,
					G_TYPE_INVALID)) {
			g_printerr ("Error while retrieving file destination: %s\n", 
				    error->message);
			g_error_free (error);
			return 0;
		}
		klass = g_type_class_ref (MALEM_TYPE_FILE_TRANSFER_DIRECTION);
		evalue = g_enum_get_value_by_nick (klass, dir);
		direction = evalue->value;
		g_type_class_unref (klass);
		return direction;
	} else {
		return 0;
	}
}

static gchar **
impl_get_files_for_directory (MalemFileTransfer  *transfer,
			      const gchar        *directory)
{
	DBusGProxy *proxy;
	gchar **files;
	GError *error = NULL;
	
	proxy = get_proxy (transfer);
	
	if (proxy != NULL) {
		if (!dbus_g_proxy_call (proxy, "GetFilesForDirectory", &error,
					G_TYPE_STRING, directory,
					G_TYPE_INVALID,
					G_TYPE_STRV, &files,
					G_TYPE_INVALID)) {
			g_printerr ("Error while retrieving file list: %s\n", 
				    error->message);
			g_error_free (error);
			return NULL;
		}
		return files;
	} else {
		return NULL;
	}
}

void
malem_file_transfer_init_default (MalemFileTransfer *transfer)
{
	get_proxy (transfer);
}

void
malem_file_transfer_iface_init_default (MalemFileTransferIface *iface)
{
	iface->add_file = impl_add_file;
	iface->get_file_direction = impl_get_file_direction;
	iface->get_files_for_directory = impl_get_files_for_directory;
}
