/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <glib.h>
#include <string.h>
#include "malem-file-transfer.h"

#define HASH_TABLE_DATA_KEY "malem-file-transfer-hash-table"

static GHashTable *
get_hash_table (MalemFileTransfer *transfer)
{
	GHashTable *table;
	
	table = (GHashTable *) g_object_get_data (G_OBJECT (transfer),
						  HASH_TABLE_DATA_KEY);

	if (table == NULL) {
		table = g_hash_table_new_full (g_str_hash, g_str_equal,
					       (GDestroyNotify) g_free,
					       NULL);
		g_object_set_data_full (G_OBJECT (transfer),
					HASH_TABLE_DATA_KEY,
					(gpointer) table,
					(GDestroyNotify) g_hash_table_unref);
	}

	return table;
}

static void
impl_add_file (MalemFileTransfer          *transfer,
	       const gchar                *filename,
	       MalemFileTransferDirection  direction)
{
	GHashTable *table;
	MalemFileTransferDirection old_direction;

	table = get_hash_table (transfer);
	old_direction = GPOINTER_TO_UINT (g_hash_table_lookup (table,
							       filename));

	if (old_direction == direction)
		return;

	g_hash_table_replace (table,
			      (gpointer) g_strdup (filename),
			      GUINT_TO_POINTER (old_direction | direction));

	if (old_direction == 0)
		malem_file_transfer_file_added (transfer, filename);
}

static MalemFileTransferDirection
impl_get_file_direction (MalemFileTransfer *transfer,
			 const gchar       *filename)
{
	GHashTable *table = get_hash_table (transfer);
	MalemFileTransferDirection direction =
		GPOINTER_TO_UINT (g_hash_table_lookup (table, filename));
	return direction;
}

static void
files_from_dir_func (const gchar *key,
		     gpointer     value,
		     gpointer     user_data[2])
{
	GPatternSpec *pspec = *((GPatternSpec **) user_data[0]);
	GArray **files = (GArray **) user_data[1];
	
	if (g_pattern_match_string (pspec, key)) {
		gchar *key_copy = g_strdup (key);
		g_array_append_val (*files, key_copy);
	}
}

static gchar **
impl_get_files_for_directory (MalemFileTransfer *transfer,
			      const gchar       *directory)
{
	GHashTable *table;
	gint len;
	gchar *pattern;
	GPatternSpec *pspec;
	GArray *files;
	gpointer user_data[2] = { &pspec, &files };
	
	len = strlen(directory);
	
	if (len == 0)
		return NULL;
	else if (directory[len-1] != '/')
		pattern = g_strdup_printf ("%s/*", directory);
	else
		pattern = g_strdup_printf ("%s*", directory);
	pspec = g_pattern_spec_new (pattern);
	g_free (pattern);
	
	table = get_hash_table (transfer);

	files = g_array_new (TRUE, FALSE, sizeof (gchar *));
	g_hash_table_foreach (table, (GHFunc) files_from_dir_func, user_data);
	
	g_pattern_spec_free (pspec);
	return (gchar **) g_array_free (files, FALSE);
}


void
malem_file_transfer_init_default (MalemFileTransfer *transfer)
{
}

void
malem_file_transfer_iface_init_default (MalemFileTransferIface *iface)
{
	iface->add_file = impl_add_file;
	iface->get_file_direction = impl_get_file_direction;
	iface->get_files_for_directory = impl_get_files_for_directory;
}
