/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif
#include "malem-file-transfer.h"
#include "malem-task.h"
#include <string.h>

enum {
	FILE_ADDED,
	LAST_SIGNAL
};

static guint transfer_signals[LAST_SIGNAL];

enum {
	MALEM_FILE_TRANSFER_ERROR_INVALID_DIRECTION,
};

static GQuark
malem_file_transfer_error_quark (void)
{
	static GQuark quark = 0;
	if (G_UNLIKELY (quark == 0))
		quark = g_quark_from_static_string ("malem-file-transfer-error-quark");
	return quark;
}

static gboolean
malem_file_transfer_dbus_get_file_direction (MalemFileTransfer  *transfer,
					     gchar              *filename,
					     gchar             **direction,
					     GError            **error)
{
	GEnumClass *klass;
	GEnumValue *evalue;
	MalemFileTransferDirection dir;
	
	dir = malem_file_transfer_get_file_direction (transfer, filename);
	
	klass = g_type_class_ref (MALEM_TYPE_FILE_TRANSFER_DIRECTION);
	evalue = g_enum_get_value (klass, dir);
	*direction = g_strdup (evalue->value_nick);
	g_type_class_unref (klass);
	return TRUE;
}

static gboolean
malem_file_transfer_dbus_add_file (MalemFileTransfer  *transfer,
				   gchar              *filename,
				   gchar              *direction,
				   GError            **error)
{
	GEnumClass *klass;
	GEnumValue *evalue;

	klass = g_type_class_ref (MALEM_TYPE_FILE_TRANSFER_DIRECTION);
	evalue = g_enum_get_value_by_nick (klass, direction);

	if (evalue == NULL) {
		g_set_error (error,
			     malem_file_transfer_error_quark (),
			     MALEM_FILE_TRANSFER_ERROR_INVALID_DIRECTION,
			     "Invalid direction: %s",
			     direction);
		return FALSE;
	}

	malem_file_transfer_add_file (transfer, filename, evalue->value);
	g_type_class_unref (klass);
	return TRUE;
}

static gboolean
malem_file_transfer_dbus_get_files_for_directory (MalemFileTransfer   *transfer,
						  gchar               *directory,
						  gchar             ***files,
						  GError             **error)
{
	*files = malem_file_transfer_get_files_for_directory (transfer,
							      directory);
	return TRUE;
}
#include "malem-file-transfer-dbus-glue.h"

static void
malem_file_transfer_class_init (gpointer g_iface)
{
	GType iface_type = G_TYPE_FROM_INTERFACE (g_iface);

	transfer_signals[FILE_ADDED] =
		g_signal_new ("file-added",
			      iface_type,
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (MalemFileTransferIface, file_added),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE,
			      1, G_TYPE_STRING);

	dbus_g_object_type_install_info (iface_type,
					 &dbus_glib_malem_file_transfer_object_info);
}

GType
malem_file_transfer_get_type (void)
{
	static GType the_type = 0;
	
	if (G_UNLIKELY (the_type == 0)) {
		static const GTypeInfo info = {
			sizeof (MalemFileTransferIface),
			NULL, NULL,
			(GClassInitFunc) malem_file_transfer_class_init,
			NULL, NULL, 0, 0, NULL
		};
		
		the_type = g_type_register_static (G_TYPE_INTERFACE,
						   g_intern_static_string ("MalemFileTransfer"),
						   &info, 0);
		g_type_interface_add_prerequisite (the_type, MALEM_TYPE_TASK);
	}
	return the_type;
}

void
malem_file_transfer_add_file (MalemFileTransfer          *transfer,
			      const gchar                *filename,
			      MalemFileTransferDirection  direction)
{
	g_return_if_fail (MALEM_IS_FILE_TRANSFER (transfer));
	g_return_if_fail (filename != NULL);

	(* MALEM_FILE_TRANSFER_GET_IFACE (transfer)->add_file) (transfer,
								filename,
								direction);
}

MalemFileTransferDirection
malem_file_transfer_get_file_direction (MalemFileTransfer *transfer,
					const gchar       *filename)
{
	g_return_val_if_fail (MALEM_IS_FILE_TRANSFER (transfer), 0);
	g_return_val_if_fail (filename != NULL, 0);

	return (* MALEM_FILE_TRANSFER_GET_IFACE (transfer)->get_file_direction) 
							(transfer, filename);
}

gchar 	**
malem_file_transfer_get_files_for_directory (MalemFileTransfer *transfer,
					     const gchar       *directory)
{
	g_return_val_if_fail (MALEM_IS_FILE_TRANSFER (transfer), 0);
	g_return_val_if_fail (directory != NULL, 0);

	return (* MALEM_FILE_TRANSFER_GET_IFACE (transfer)->get_files_for_directory)
							(transfer, directory);
}

void
malem_file_transfer_file_added (MalemFileTransfer *transfer,
				const gchar       *filename)
{
	g_return_if_fail (MALEM_IS_FILE_TRANSFER (transfer));

	g_signal_emit (transfer, transfer_signals[FILE_ADDED], 0, filename);
}
