/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "malem-interfaces.h"

#include "malem-dictionary.h"
#include "malem-file-transfer.h"
#include "malem-frobnicator.h"

static void
malem_interfaces_init (void)
{	
	malem_register_interface_static ("org.gnome.Mathusalem.Dictionary",
					 MALEM_TYPE_DICTIONARY,
					 (GInterfaceInitFunc) malem_dictionary_iface_init_default,
					 (GInstanceInitFunc) malem_dictionary_init_default);
	malem_register_interface_static ("org.gnome.Mathusalem.FileTransfer",
					 MALEM_TYPE_FILE_TRANSFER,
					 (GInterfaceInitFunc) malem_file_transfer_iface_init_default,
					 (GInstanceInitFunc) malem_file_transfer_init_default);
	malem_register_interface_static ("org.gnome.Mathusalem.Frobnicator",
					 MALEM_TYPE_FROBNICATOR,
					 (GInterfaceInitFunc) malem_frobnicator_iface_init_default,
					 (GInstanceInitFunc) malem_frobnicator_init_default);
}

/* -------------------------------------------------------------------------- */

static GHashTable *interfaces = NULL;

typedef struct {
	GType               iface_type;
	GInterfaceInitFunc  interface_init;
	GInstanceInitFunc   instance_init;
} MalemInterfaceInfo;

static GQuark
iface_name_quark (void)
{
	static GQuark quark = 0;
	if (quark == 0)
		quark = g_quark_from_static_string ("dbus-interface-name");
	return quark;
}

void
malem_register_interface_static (const char         *iface_name,
				 GType               iface_type,
				 GInterfaceInitFunc  iface_init,
				 GInstanceInitFunc   instance_init)
{
	MalemInterfaceInfo *iface_info;
	
	iface_info = g_new (MalemInterfaceInfo, 1);
	iface_info->iface_type = iface_type;
	iface_info->interface_init = iface_init;
	iface_info->instance_init = instance_init;
	
	g_type_set_qdata (iface_type, 
			  iface_name_quark (),
			  (gpointer) iface_name);
	
	g_hash_table_insert (interfaces, 
			     GUINT_TO_POINTER (g_quark_from_static_string (iface_name)),
			     (gpointer) iface_info);
}

static MalemInterfaceInfo *
lookup_interface_info (GQuark iface_quark)
{
	if (G_UNLIKELY (interfaces == NULL)) {
		interfaces = g_hash_table_new (g_direct_hash, g_direct_equal);
		malem_interfaces_init ();
	}

	return g_hash_table_lookup (interfaces, GUINT_TO_POINTER (iface_quark));
}

GType
malem_interface_lookup (GQuark iface_quark)
{
	MalemInterfaceInfo *iface_info = lookup_interface_info (iface_quark);
	
	g_return_val_if_fail (iface_info != NULL, 0);
		
	return iface_info->iface_type;
}

void
malem_type_add_interface_static (GType  instance_type,
				 GQuark iface_quark)
{
	GInterfaceInfo info = { NULL, NULL, NULL };
	MalemInterfaceInfo *iface_info = lookup_interface_info (iface_quark);
	
	g_return_if_fail (iface_info != NULL);

	info.interface_init = iface_info->interface_init;

	g_type_add_interface_static (instance_type,
				     iface_info->iface_type,
				     &info);
}

void
malem_object_init_interfaces (GObject *object)
{
	GObjectClass *klass;
	GType *interfaces;
	const gchar *name;
	MalemInterfaceInfo *info;
	guint i;

	g_return_if_fail (G_IS_OBJECT (object));

	klass = G_OBJECT_GET_CLASS (object);

	interfaces = g_type_interfaces (G_TYPE_FROM_CLASS (klass), NULL);
	
	for (i = 0; interfaces[i] != G_TYPE_INVALID; i++)
	{	
		name = g_type_get_qdata (interfaces[i], iface_name_quark ());

		if (name == NULL)
			continue;

		info = lookup_interface_info (g_quark_from_string (name));
		info->instance_init ((GTypeInstance *) object, klass);
	}
	g_free (interfaces);
}
