/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif
#include "malem-task.h"
#include "malem-enum-types.h"
#include "interfaces/malem-interfaces.h"

enum {
	PROGRESS_CHANGED,
	STATUS_CHANGED,
	ACTION_TRIGGERED,

	LAST_SIGNAL
};

static guint task_signals[LAST_SIGNAL];

static void malem_task_class_init (gpointer g_iface);

static gboolean	malem_task_dbus_get_status		(MalemTask *task, gchar **status, GError **error);
static gboolean malem_task_dbus_start			(MalemTask *task, GError **error);
static gboolean malem_task_dbus_pause			(MalemTask *task, GError **error);
static gboolean malem_task_dbus_abort			(MalemTask *task, GError **error);
static gboolean malem_task_dbus_set_progress		(MalemTask *task, guint progress, GError **error);
static gboolean malem_task_dbus_get_progress		(MalemTask *task, guint *progress, GError **error);
static gboolean malem_task_dbus_get_title		(MalemTask *task, gchar **title, GError **error);
static gboolean malem_task_dbus_get_owner		(MalemTask *task, gchar **owner, GError **error);
static gboolean malem_task_dbus_add_action		(MalemTask *task, const gchar *action_name, const gchar *label,
							 const gchar *icon_name, GError **error);
static gboolean	malem_task_dbus_has_action		(MalemTask *task, const gchar *action_name,
							 gboolean *has_action, GError **error);
static gboolean malem_task_dbus_list_actions		(MalemTask *task, gchar ***actions, GError **error);
static gboolean malem_task_dbus_trigger_action		(MalemTask *task, const gchar *action_name, GError **error);
static gboolean malem_task_dbus_implements_interface	(MalemTask *task, const gchar *interface_name,
							 gboolean *implemented, GError **error);
static gboolean malem_task_dbus_list_interfaces		(MalemTask *task, gchar ***interfaces, GError **error);
#include "malem-task-dbus-glue.h"

GType
malem_task_get_type (void)
{
	static GType the_type = 0;
	
	if (G_UNLIKELY (the_type == 0)) {
		static GTypeInfo type_info = {
			sizeof (MalemTaskIface),
			NULL, NULL,
			(GClassInitFunc) malem_task_class_init
		};
		the_type = g_type_register_static (G_TYPE_INTERFACE,
						   g_intern_static_string ("MalemTask"),
						   &type_info, 0);
		g_type_interface_add_prerequisite (the_type, G_TYPE_OBJECT);
	}
	return the_type;
}

static void
malem_task_class_init (gpointer g_iface)		
{
	GType iface_type = G_TYPE_FROM_INTERFACE (g_iface);

	task_signals[PROGRESS_CHANGED] =
		g_signal_new ("progress-changed",
			      iface_type,
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (MalemTaskIface, progress_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__UINT,
			      G_TYPE_NONE,
			      1, G_TYPE_UINT);

	task_signals[STATUS_CHANGED] =
		g_signal_new ("status-changed",
			      iface_type,
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (MalemTaskIface, status_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__ENUM,
			      G_TYPE_NONE,
			      1, /*MALEM_TYPE_TASK_STATUS*/ G_TYPE_UINT);

	task_signals[ACTION_TRIGGERED] =
		g_signal_new ("action-triggered",
			      iface_type,
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (MalemTaskIface, action_triggered),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE,
			      1, G_TYPE_STRING);

	g_object_interface_install_property (g_iface,
					     g_param_spec_string ("title",
					 			  "Title",
							 	  "The task title",
							 	  NULL,
							 	  G_PARAM_READWRITE |
							 	  G_PARAM_CONSTRUCT_ONLY));
	g_object_interface_install_property (g_iface,
					     g_param_spec_string ("owner",
					 			  "Owner",
							 	  "The task owner application",
							 	  NULL,
							 	  G_PARAM_READWRITE |
							 	  G_PARAM_CONSTRUCT_ONLY));
	g_object_interface_install_property (g_iface,
					     g_param_spec_uint ("progress",
								"Progress",
							 	"The task progress status",
							 	0, 100, 0,
							 	G_PARAM_READWRITE));
	g_object_interface_install_property (g_iface,
					     g_param_spec_enum ("status",
					     			"Status",
					     			"The task status",
								MALEM_TYPE_TASK_STATUS,
					     			MALEM_TASK_STATUS_NEW,
					     			G_PARAM_READWRITE));

	dbus_g_object_type_install_info (iface_type,
					 &dbus_glib_malem_task_object_info);
}

/* Status */

MalemTaskStatus
malem_task_get_status (MalemTask *task)
{
	MalemTaskStatus status;
	g_return_val_if_fail (MALEM_IS_TASK (task), 0);
	g_object_get (task, "status", &status, NULL);
	return status;
}

void
malem_task_start (MalemTask *task)
{
	g_return_if_fail (MALEM_IS_TASK (task));
	(* MALEM_TASK_GET_IFACE (task)->start) (task);
}

void
malem_task_pause (MalemTask *task)
{
	g_return_if_fail (MALEM_IS_TASK (task));
	(* MALEM_TASK_GET_IFACE (task)->pause) (task);
}

void
malem_task_abort (MalemTask *task)
{
	g_return_if_fail (MALEM_IS_TASK (task));
	(* MALEM_TASK_GET_IFACE (task)->abort) (task);
}

void
malem_task_status_changed (MalemTask       *task,
			   MalemTaskStatus  status)
{
	g_return_if_fail (MALEM_IS_TASK (task));
	g_signal_emit (task, task_signals[STATUS_CHANGED], 0, status);
}

/* Progression */ 

guint
malem_task_get_progress (MalemTask *task)
{
	guint progress;
	g_return_val_if_fail (MALEM_IS_TASK (task), 0);
	g_object_get (task, "progress", &progress, NULL);
	return progress;
}

void
malem_task_set_progress (MalemTask *task,
			 guint      progress)
{
	g_return_if_fail (MALEM_IS_TASK (task));
	g_object_set (task, "progress", progress, NULL);
}

void
malem_task_progress_changed (MalemTask *task,
			     guint      progress)
{
	g_return_if_fail (MALEM_IS_TASK (task));
	g_signal_emit (task, task_signals[PROGRESS_CHANGED], 0, progress);
}

/* Properties */

gchar *
malem_task_get_title (MalemTask *task)
{
	gchar *title;
	g_return_val_if_fail (MALEM_IS_TASK (task), NULL);
	g_object_get (task, "title", &title, NULL);
	return title;
}

gchar *
malem_task_get_owner (MalemTask *task)
{
	gchar *owner;
	g_return_val_if_fail (MALEM_IS_TASK (task), NULL);
	g_object_get (task, "owner", &owner, NULL);
	return owner;
}

/* Custom actions */

void
malem_task_add_action (MalemTask   *task,
		       const gchar *action_name,
		       const gchar *label,
		       const gchar *icon_name)
{
	g_return_if_fail (MALEM_IS_TASK (task));
	g_return_if_fail (action_name != NULL);
	(* MALEM_TASK_GET_IFACE (task)->add_action) (task, action_name, label, icon_name);
}

gboolean
malem_task_has_action (MalemTask   *task,
		       const gchar *action_name)
{
	g_return_val_if_fail (MALEM_IS_TASK (task), FALSE);
	return (* MALEM_TASK_GET_IFACE (task)->has_action) (task, action_name);
}

gchar **
malem_task_list_actions (MalemTask   *task)
{
	g_return_val_if_fail (MALEM_IS_TASK (task), NULL);
	return (* MALEM_TASK_GET_IFACE (task)->list_actions) (task);
}

void
malem_task_trigger_action (MalemTask   *task,
			   const gchar *action_name)
{
	g_return_if_fail (MALEM_IS_TASK (task));
	(* MALEM_TASK_GET_IFACE (task)->trigger_action) (task, action_name);
}

void
malem_task_action_triggered (MalemTask   *task,
			     const gchar *action_name)
{
	g_return_if_fail (MALEM_IS_TASK (task));
	g_signal_emit (task, task_signals[ACTION_TRIGGERED], 0, action_name);
}

/* Custom interfaces */

gboolean
malem_task_implements_interface (MalemTask *task,
				 GType      iface)
{
	g_return_val_if_fail (MALEM_IS_TASK (task), FALSE);
	return (* MALEM_TASK_GET_IFACE (task)->implements_interface) (task, iface);
}

GType *
malem_task_list_interfaces (MalemTask *task)
{
	g_return_val_if_fail (MALEM_IS_TASK (task), FALSE);
	return (* MALEM_TASK_GET_IFACE (task)->list_interfaces) (task);
}

/* DBus */

static gboolean
malem_task_dbus_get_status (MalemTask  *task,
			    gchar     **status,
			    GError    **error)
{
	GEnumValue *evalue;

	evalue = g_enum_get_value (g_type_class_peek (MALEM_TYPE_TASK_STATUS),
				   malem_task_get_status (task));
	*status = g_strdup (evalue->value_nick);
	return TRUE;
}

static gboolean
malem_task_dbus_start (MalemTask  *task,
		       GError    **error)
{
	malem_task_start (task);
	return TRUE;
}
static gboolean
malem_task_dbus_pause (MalemTask  *task,
		       GError    **error)
{
	malem_task_pause (task);
	return TRUE;
}

static gboolean
malem_task_dbus_abort (MalemTask  *task,
		       GError    **error)
{
	malem_task_abort (task);
	return TRUE;
}

static gboolean
malem_task_dbus_set_progress (MalemTask  *task,
			      guint       progress,
			      GError    **error)
{
	malem_task_set_progress (task, progress);
	return TRUE;
}

static gboolean
malem_task_dbus_get_progress (MalemTask  *task,
			      guint      *progress,
			      GError    **error)
{
	*progress = malem_task_get_progress (task);
	return TRUE;
}

static gboolean
malem_task_dbus_get_title (MalemTask  *task,
			   gchar     **title,
			   GError    **error)
{
	*title = malem_task_get_title (task);
	return TRUE;
}

static gboolean
malem_task_dbus_get_owner (MalemTask  *task,
			   gchar     **owner,
			   GError    **error)
{
	*owner = malem_task_get_owner (task);
	return TRUE;
}

static gboolean
malem_task_dbus_add_action (MalemTask    *task,
			    const gchar  *action_name,
			    const gchar  *label,
			    const gchar  *icon_name,
			    GError     **error)
{
	malem_task_add_action (task, action_name, label, icon_name);
	return TRUE;
}

static gboolean
malem_task_dbus_has_action (MalemTask    *task,
			    const gchar  *action_name,
			    gboolean     *has_action,
			    GError      **error)
{
	*has_action = malem_task_has_action (task, action_name);
	return TRUE;
}

static gboolean
malem_task_dbus_list_actions (MalemTask   *task,
			      gchar     ***actions,
			      GError     **error)
{
	*actions = malem_task_list_actions (task);
	return TRUE;
}

static gboolean
malem_task_dbus_trigger_action (MalemTask    *task,
				const gchar  *action_name,
				GError      **error)
{
	malem_task_trigger_action (task, action_name);
	return TRUE;
}

static gboolean
malem_task_dbus_implements_interface (MalemTask    *task,
				      const gchar  *iface_name,
				      gboolean     *implemented,
				      GError      **error)
{
	GType iface;
	GQuark iface_quark;
	
	iface_quark = g_quark_try_string (iface_name);
	iface = malem_interface_lookup (iface_quark);
	
	if (iface == 0)
		*implemented = FALSE;
	else
		*implemented = malem_task_implements_interface (task, iface);
	return TRUE;
}

static gboolean
malem_task_dbus_list_interfaces (MalemTask   *task,
				 gchar     ***interfaces,
				 GError     **error)
{
	GQuark quark;
	GType *ifaces, *p;
	GArray *array;
	gchar *iface_name;

	quark = g_quark_from_static_string ("dbus-interface-name");

	array = g_array_new (TRUE, FALSE, sizeof (gchar *));

	ifaces = malem_task_list_interfaces (task);	

	g_assert (ifaces != NULL);

	for (p = ifaces; *p != 0; p++) {
		iface_name = g_type_get_qdata (*p, quark);
		if (iface_name != NULL) {
			gchar *val = g_strdup (iface_name);
			g_array_append_val (array, val);
		}
	}

	g_free (ifaces);
	*interfaces = (gchar **) g_array_free (array, FALSE);
	return TRUE;
}
