/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <dbus/dbus-protocol.h>
#include <dbus/dbus-glib.h>
#include <gtk/gtk.h>
#include <libnotify/notify.h>

#include "malem-app.h"
#include "malem-task-default.h"
#include "malem-task-list.h"
#include "malem-debug.h"
#include "interfaces/malem-interfaces.h"

#include "gui/malem-tray-icon.h"
#include "gui/malem-progress-window.h"

#define FACTORY_NAMESPACE	"org.gnome.Mathusalem"
#define FACTORY_OBJECT_PATH	"/org/gnome/Mathusalem"
#define TASK_OBJECT_PATH	"/org/gnome/Mathusalem/Tasks/%08x"

#define MALEM_APP_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o),	\
				  MALEM_TYPE_APP,			\
				  MalemAppPrivate))

struct _MalemAppPrivate {
	DBusGConnection *connection;
	MalemTaskList   *tasks;
	GtkWidget       *window;
	MalemTrayIcon   *icon;
};

static void		 malem_factory_iface_init	(MalemFactoryIface *iface);
static MalemTask	*malem_app_register_task	(MalemFactory  *factory,
							 const gchar  **interfaces,
							 const gchar   *title,
							 const gchar   *owner);
static MalemTask	**malem_app_list_tasks		(MalemFactory  *factory);
static MalemTask	**malem_app_list_tasks_for_interface
							(MalemFactory  *factory,
							 const gchar   *interface);
static void		 malem_app_get_version		(MalemFactory  *factory,
							 gint          *major,
							 gint          *minor,
							 gint          *micro);

G_DEFINE_TYPE_WITH_CODE (MalemApp, malem_app, G_TYPE_OBJECT,
			 G_IMPLEMENT_INTERFACE (MALEM_TYPE_FACTORY,
				 		malem_factory_iface_init))

static void
on_tray_icon_quit (MalemTrayIcon *icon,
		   gpointer       user_data)
{
	gtk_main_quit ();
}

static void
on_tray_icon_clicked (MalemTrayIcon *icon,
		      gpointer       user_data)
{
	MalemApp *app = MALEM_APP (user_data);
	
	if (app->priv->window == NULL) {
		app->priv->window = malem_progress_window_new (app->priv->tasks,
							       GTK_STATUS_ICON (icon));
		g_signal_connect (app->priv->window, "hide",
				  G_CALLBACK (gtk_widget_destroy), NULL);
		g_object_add_weak_pointer (G_OBJECT (app->priv->window),
					   (gpointer) &app->priv->window);
		gtk_widget_show (app->priv->window);
	} else {
		gtk_widget_destroy (app->priv->window);
	}
}

static void
on_task_added (MalemTaskList     *tasklist,
	       MalemTaskListIter *iter,
	       MalemApp          *app)
{
	gboolean visible = malem_task_list_get_length (app->priv->tasks) != 0;
	gtk_status_icon_set_visible (GTK_STATUS_ICON (app->priv->icon), visible);
}

static void
on_notification_closed (NotifyNotification *notification,
		        MalemApp           *app)
{
	gboolean visible = malem_task_list_get_length (app->priv->tasks) != 0;

	g_object_unref (notification);

	if (!visible && app->priv->window != NULL)
		gtk_widget_destroy (app->priv->window);

	gtk_status_icon_set_visible (GTK_STATUS_ICON (app->priv->icon), visible);
}

static void
notify (MalemTask *task,
	MalemApp  *app)
{
	NotifyNotification *notification;
	MalemTaskStatus status;
	const gchar *summary, *body, *icon;
	
	status = malem_task_get_status (task);
	
	if (status == MALEM_TASK_STATUS_COMPLETED) {
		summary = "Task completed";
		icon = "gtk-apply";
	} else if (status == MALEM_TASK_STATUS_ABORTED) {
		summary = "Task aborted";
		icon = "gtk-stop";
	} else {
		return;
	}

	body = malem_task_get_title (task);

	notification = notify_notification_new (summary, body, icon, NULL);
	g_signal_connect (notification, "closed",
			  G_CALLBACK (on_notification_closed), app);
	notify_notification_attach_to_status_icon (notification,
						   GTK_STATUS_ICON (app->priv->icon));
	if (!notify_notification_show (notification, NULL))
		on_notification_closed (notification, app);
}

static void
on_task_removed (MalemTaskList     *tasklist,
		 MalemTaskListIter *iter,
		 MalemApp          *app)
{
	/* Note: before the task is actually removed -> count == 1 */
	gboolean visible = malem_task_list_get_length (app->priv->tasks) != 1;

	if (!visible && app->priv->window != NULL)
		gtk_widget_destroy (app->priv->window);
	
	notify (malem_task_list_iter_get_task (iter), app);
}


static void
malem_app_init (MalemApp *app)
{
	app->priv = MALEM_APP_GET_PRIVATE (app);

	app->priv->tasks = malem_task_list_new ();
	g_object_ref_sink (app->priv->tasks);

	app->priv->icon = malem_tray_icon_new ();
	g_signal_connect (app->priv->icon, "quit",
			  G_CALLBACK (on_tray_icon_quit), NULL);
	g_signal_connect (app->priv->icon, "activate",
			  G_CALLBACK (on_tray_icon_clicked), app);
	gtk_status_icon_set_visible (GTK_STATUS_ICON (app->priv->icon), FALSE);
	
	g_signal_connect (app->priv->tasks, "task-added",
			  G_CALLBACK (on_task_added), app);
	g_signal_connect (app->priv->tasks, "task-removed",
			  G_CALLBACK (on_task_removed), app);

	malem_debug ("app", "MalemApp Initialized");
}

static void
malem_app_finalize (GObject *object)
{
	MalemApp *app = MALEM_APP (object);

	g_return_if_fail (app->priv != NULL);

	dbus_g_connection_unref (app->priv->connection);
	
	g_object_unref (app->priv->tasks);
	g_object_unref (app->priv->icon);
	if (app->priv->window)
		g_object_unref (app->priv->window);

	malem_debug ("app", "MalemApp Finalized");

	/* chainup */
	G_OBJECT_CLASS (malem_app_parent_class)->finalize (object);
}

static void
malem_factory_iface_init (MalemFactoryIface *iface)
{
	iface->register_task = malem_app_register_task;
	iface->list_tasks = malem_app_list_tasks;
	iface->list_tasks_for_interface = malem_app_list_tasks_for_interface;
	iface->get_version = malem_app_get_version;
}

static void
malem_app_class_init (MalemAppClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	
	object_class->finalize = malem_app_finalize;

	g_type_class_add_private (klass, sizeof (MalemAppPrivate));
}

static void
app_weak_notify (gpointer data,
		 GObject *where_the_app_was)
{
	gtk_main_quit ();
}

MalemApp *
malem_app_get_default (void)
{
	static MalemApp *app = NULL;
	
	if (G_UNLIKELY (app == NULL)) {
		DBusGConnection *connection;
		DBusGProxy *proxy;
		GError *error = NULL;
		guint32 ret;

		/* Get the connection and ensure the name is not used yet */
		connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
		if (connection == NULL) {
			g_warning ("Failed to make connection to session bus: %s",
				   error->message);
			g_error_free (error);
			return NULL;
		}
		
		proxy = dbus_g_proxy_new_for_name (connection, DBUS_SERVICE_DBUS,
						   DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
	
	        if (!dbus_g_proxy_call (proxy, "RequestName", &error,
			                G_TYPE_STRING, FACTORY_NAMESPACE,
			                G_TYPE_UINT, 0,
			                G_TYPE_INVALID,
			                G_TYPE_UINT, &ret,
			                G_TYPE_INVALID)) {
			g_warning ("There was an error requesting the name: %s",
				   error->message);
			g_error_free (error);
			return NULL;
		}
		
		if (ret != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {
			/* Someone else registered the name before us */
			return NULL;
		}
		
		/* Invoke parent constructor. */
		app = MALEM_APP (g_object_new (MALEM_TYPE_APP, NULL));
		
		/* Register the app on the bus */
		dbus_g_connection_register_g_object (connection,
						     FACTORY_OBJECT_PATH,
						     G_OBJECT (app));

		dbus_g_connection_ref (connection);
		app->priv->connection = connection;

		/* Monitor the object status */
		g_object_add_weak_pointer (G_OBJECT (app), (gpointer) &app);
		g_object_weak_ref (G_OBJECT (app), app_weak_notify, NULL);
	}
	
	return app;
}

static MalemTask *
malem_app_register_task (MalemFactory  *factory,
			 const gchar  **interfaces,
			 const gchar   *title,
			 const gchar   *owner)
{
	static guint task_uniqid = 0;
	gchar *object_path;
	MalemApp *app = MALEM_APP (factory);
	MalemTask *task;

	task = MALEM_TASK (malem_task_default_new (interfaces, title, owner));
	malem_task_list_add_task (app->priv->tasks, task);

	object_path = g_strdup_printf (TASK_OBJECT_PATH, task_uniqid++);
	dbus_g_connection_register_g_object (app->priv->connection,
					     object_path,
					     G_OBJECT (task));
	g_free (object_path);

	malem_factory_task_registered (factory, task, interfaces);

	return task;
}

static MalemTask **
malem_app_list_tasks (MalemFactory  *factory)
{
	return malem_app_list_tasks_for_interface (factory, NULL);
}

static MalemTask **
malem_app_list_tasks_for_interface (MalemFactory  *factory,
				    const gchar   *interface)
{
	MalemTaskList *tasklist;
	gint length;
	GPtrArray *tasks;
	MalemTaskListIter iter;
	MalemTask *task;
	GType iface_type = G_TYPE_INVALID;

	tasklist = MALEM_APP (factory)->priv->tasks;
	
	length = malem_task_list_get_length (tasklist);
	
	tasks = g_ptr_array_sized_new (length + 1);
	
	if (length > 0) {
		malem_task_list_get_iter_first (tasklist, &iter);
		
		if (interface != NULL)
			iface_type = malem_interface_lookup (g_quark_from_string (interface));
		
		do {
			task = malem_task_list_iter_get_task (&iter);
			if (interface == NULL || 
			    malem_task_implements_interface (task, iface_type))
				g_ptr_array_add (tasks, task);
		} while (malem_task_list_iter_next (&iter));
	}
	
	g_ptr_array_add (tasks, NULL);
	return (MalemTask **) g_ptr_array_free (tasks, FALSE);
}

static void
malem_app_get_version (MalemFactory  *factory,
		       gint          *major,
		       gint          *minor,
		       gint          *micro)
{
	*major = MALEM_MAJOR_VERSION;
	*minor = MALEM_MINOR_VERSION;
	*micro = MALEM_MICRO_VERSION;
}
