/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "malem-task-default.h"
#include "malem-debug.h"

#include "interfaces/malem-interfaces.h"

#define MALEM_TASK_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o),	\
				   MALEM_TYPE_TASK_DEFAULT,		\
				   MalemTaskDefaultPrivate))

typedef struct _MalemTaskDefaultAction {
	gchar *name;
	gchar *label;
	gchar *icon_name;
} MalemTaskDefaultAction;

struct _MalemTaskDefaultPrivate {
	MalemTaskStatus   status;
	gchar            *title;
	guint             progress;
	gchar            *owner;
	GList            *actions;
};

enum {
	PROP_0,
	PROP_STATUS,
	PROP_TITLE,
	PROP_PROGRESS,
	PROP_OWNER
};

static gpointer malem_task_default_parent_class = NULL;

static void
malem_task_default_init (MalemTaskDefault *task)
{
	task->priv = MALEM_TASK_GET_PRIVATE (task);

	malem_debug ("task", "MalemTaskDefault Initialized");
}

static void
malem_task_default_finalize (GObject *object)
{
	MalemTaskDefault *task = MALEM_TASK_DEFAULT (object);
	GList *p, *l;

	malem_debug ("task", "MalemTaskDefault Finalized");

	g_free (task->priv->title);
	g_free (task->priv->owner);

	p = task->priv->actions;
	while (p != NULL) {
		l = p;
		p = l->next;
		g_free (l->data);
		g_list_free_1 (l);
	}

	/* chainup */
	G_OBJECT_CLASS (malem_task_default_parent_class)->finalize (object);
}

static void
malem_task_default_get_property (GObject    *object,
				 guint       prop_id,
				 GValue     *value,
				 GParamSpec *pspec)
{
	MalemTaskDefault *task = MALEM_TASK_DEFAULT (object);

	switch (prop_id)
	{
		case PROP_STATUS:
			g_value_set_enum (value, task->priv->status);
			break;
		case PROP_TITLE:
			g_value_set_string (value, task->priv->title);
			break;
		case PROP_PROGRESS:
			g_value_set_uint (value, task->priv->progress);
			break;
		case PROP_OWNER:
			g_value_set_string (value, task->priv->owner);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
malem_task_default_set_property (GObject      *object,
				 guint         prop_id,
				 const GValue *value,
				 GParamSpec   *pspec)
{
	MalemTaskDefault *task = MALEM_TASK_DEFAULT (object);

	switch (prop_id)
	{
		case PROP_STATUS:
			task->priv->status = g_value_get_enum (value);
			malem_task_status_changed (MALEM_TASK (object),
						   task->priv->status);
			break;
		case PROP_TITLE:
			task->priv->title = g_strdup (g_value_get_string (value));
			break;
		case PROP_PROGRESS:
			task->priv->progress = g_value_get_uint (value);
			malem_task_progress_changed (MALEM_TASK (object),
						     task->priv->progress);
			break;
		case PROP_OWNER:
			task->priv->owner = g_strdup(g_value_get_string (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
malem_task_default_start (MalemTask *task)
{
	MalemTaskDefault *taskd = MALEM_TASK_DEFAULT (task);
	if (taskd->priv->status == MALEM_TASK_STATUS_NEW ||
	    taskd->priv->status == MALEM_TASK_STATUS_PAUSED) {
		taskd->priv->status = MALEM_TASK_STATUS_STARTED;
		malem_task_status_changed (task, MALEM_TASK_STATUS_STARTED);
	}
}

static void
malem_task_default_pause (MalemTask *task)
{
	MalemTaskDefault *taskd = MALEM_TASK_DEFAULT (task);
	if (taskd->priv->status == MALEM_TASK_STATUS_STARTED) {
		taskd->priv->status = MALEM_TASK_STATUS_PAUSED;
		malem_task_status_changed (task, MALEM_TASK_STATUS_PAUSED);
	}
}

static void
malem_task_default_abort (MalemTask *task)
{
	MalemTaskDefault *taskd = MALEM_TASK_DEFAULT (task);
	if (taskd->priv->status != MALEM_TASK_STATUS_ABORTED &&
	    taskd->priv->status != MALEM_TASK_STATUS_COMPLETED) {
		taskd->priv->status = MALEM_TASK_STATUS_ABORTED;
		malem_task_status_changed (task, MALEM_TASK_STATUS_ABORTED);
	}
}

static gboolean
malem_task_default_has_action (MalemTask   *task,
			       const gchar *action_name)
{
	MalemTaskDefault *taskd = MALEM_TASK_DEFAULT (task);
	GList *p;
	MalemTaskDefaultAction *a;
	
	for (p = taskd->priv->actions; p != NULL; p = p->next) {
		a = (MalemTaskDefaultAction *) p->data;
		if (strcmp (action_name, a->name) == 0)
			return TRUE;
	}
	return FALSE;
}

static void
malem_task_default_add_action (MalemTask   *task,
			       const gchar *action_name,
			       const gchar *label,
			       const gchar *icon_name)
{
	MalemTaskDefault *taskd;
	MalemTaskDefaultAction *a;

	if (malem_task_default_has_action (task, action_name))
		return;

	a = g_new (MalemTaskDefaultAction, 1);
	a->name = g_strdup (action_name);
	a->label = g_strdup (label);
	a->icon_name = g_strdup (icon_name);

	taskd = MALEM_TASK_DEFAULT (task);
	taskd->priv->actions = g_list_prepend (taskd->priv->actions, a);
}

static gchar **
malem_task_default_list_actions (MalemTask   *task)
{
	MalemTaskDefault *taskd = MALEM_TASK_DEFAULT (task);
	MalemTaskDefaultAction *a;
	GArray *array;
	GList *p;
	gchar *new_string;
	
	array = g_array_new (TRUE, FALSE, sizeof (gchar *));
	for (p = taskd->priv->actions; p != NULL; p = p->next) {
		a = (MalemTaskDefaultAction *) p->data;
		new_string = g_strdup (a->name);
		g_array_append_val (array, new_string);
	}
	g_array_sort (array, (GCompareFunc) strcmp);
	return (gchar **) g_array_free (array, FALSE);
}

static void
malem_task_default_trigger_action (MalemTask   *task,
				   const gchar *action_name)
{
	if (malem_task_default_has_action (task, action_name))
		malem_task_action_triggered (task, action_name);
}

static gboolean
malem_task_default_implements_interface (MalemTask *task,
					 GType      iface)
{
	MalemTaskDefaultClass *klass = MALEM_TASK_DEFAULT_GET_CLASS (task);
	return g_type_interface_peek (klass, iface) != NULL;
}

static GType *
malem_task_default_list_interfaces (MalemTask *task)
{
	return g_type_interfaces (G_OBJECT_TYPE (task), NULL);
}

static void
malem_task_default_progress_changed (MalemTask *task,
				     guint      progress)
{
	MalemTaskDefault *taskd = MALEM_TASK_DEFAULT (task);
	malem_debug ("task", "MalemTask::progress-changed(%u) signal fired", progress);

	taskd->priv->progress = progress;
	if (taskd->priv->progress >= 100) {
		taskd->priv->status = MALEM_TASK_STATUS_COMPLETED;
		malem_task_status_changed (task, MALEM_TASK_STATUS_COMPLETED);
	}
}

static void
malem_task_default_class_init (MalemTaskDefaultClass *klass)
{
	GObjectClass   *object_class = G_OBJECT_CLASS (klass);

	malem_task_default_parent_class = g_type_class_peek_parent (klass);

	object_class->finalize = malem_task_default_finalize;
	object_class->get_property = malem_task_default_get_property;
	object_class->set_property = malem_task_default_set_property;

	g_object_class_override_property (object_class, PROP_STATUS, "status");
	g_object_class_override_property (object_class, PROP_TITLE, "title");
	g_object_class_override_property (object_class, PROP_PROGRESS, "progress");
	g_object_class_override_property (object_class, PROP_OWNER, "owner");

	g_type_class_add_private (klass, sizeof (MalemTaskDefaultPrivate));
}

static void
malem_task_iface_init (MalemTaskIface *iface)
{
	iface->start = malem_task_default_start;
	iface->pause = malem_task_default_pause;
	iface->abort = malem_task_default_abort;
	iface->has_action = malem_task_default_has_action;
	iface->add_action = malem_task_default_add_action;
	iface->list_actions = malem_task_default_list_actions;
	iface->trigger_action = malem_task_default_trigger_action;
	iface->implements_interface = malem_task_default_implements_interface;
	iface->list_interfaces = malem_task_default_list_interfaces;

	iface->progress_changed = malem_task_default_progress_changed;
}

GType
malem_task_default_get_base_type (void)
{
	static GType the_type = 0;

	if (G_UNLIKELY (the_type == 0)) {
		static GTypeInfo info = {
			sizeof (MalemTaskDefaultClass),
			NULL, NULL,
			(GClassInitFunc) malem_task_default_class_init,
			NULL, NULL,
			sizeof (MalemTaskDefault), 0,
			(GInstanceInitFunc) malem_task_default_init,
			NULL
		};

		static const GInterfaceInfo iface_info = {
			(GInterfaceInitFunc) malem_task_iface_init,
			NULL, NULL 
		};

		the_type = g_type_register_static (G_TYPE_INITIALLY_UNOWNED,
						   g_intern_static_string ("MalemTaskDefault"),
						   &info,
						   (GTypeFlags) 0);

		g_type_add_interface_static (the_type,
					     MALEM_TYPE_TASK,
					     &iface_info);
	}
	return the_type;
}

static int
compare_quarks (const GQuark *q1,
		const GQuark *q2)
{
	return (*q1 > *q2) - (*q1 < *q2);
}

static gchar *
get_type_name (guint         n_ifaces, 
	       const GQuark *ifaces)
{
	guint i;
	GString *name;
	
	name = g_string_new ("MalemTaskDefault");

	for (i = 0; i < n_ifaces; i++)
		g_string_append_printf (name, "+%u", ifaces[i]);
	
	return g_string_free (name, FALSE);
}

GType
malem_task_default_get_type (const gchar *ifaces[])
{
	static GHashTable *type_table = NULL;
	GType the_type = 0;
	guint n_ifaces, i;
	GQuark *ifaces_quarks;
	gchar *type_name;

	/* There is no additional interface: return base type */
	if (ifaces == NULL || *ifaces == NULL)
		return MALEM_TYPE_TASK_DEFAULT;

	if (G_UNLIKELY (type_table == NULL))
		type_table = g_hash_table_new (g_str_hash, g_str_equal);

	n_ifaces = g_strv_length ((gchar**) ifaces);

	ifaces_quarks = g_new (GQuark, n_ifaces);
	for (i = 0; i < n_ifaces; i++)
		ifaces_quarks[i] = g_quark_from_string (ifaces[i]);
	qsort (ifaces_quarks, n_ifaces, sizeof (GQuark), (GCompareFunc) compare_quarks);

	type_name = get_type_name (n_ifaces, ifaces_quarks);

	the_type = GPOINTER_TO_UINT (
			g_hash_table_lookup (type_table, (gpointer) type_name));
		
	if (G_UNLIKELY (the_type == 0)) {
		static GTypeInfo info = {
			sizeof (MalemTaskDefaultClass),
			NULL, NULL,
			NULL, NULL, NULL,
			sizeof (MalemTaskDefault), 0,
			(GInstanceInitFunc) malem_object_init_interfaces,
			NULL
		};
		the_type = g_type_register_static (MALEM_TYPE_TASK_DEFAULT,
						   type_name,
						   &info,
						   (GTypeFlags) 0);

		for (i = 0; i < n_ifaces; i++)
			malem_type_add_interface_static (the_type,
							 ifaces_quarks[i]);
		
		g_hash_table_insert (type_table,
				     (gpointer) g_strdup (type_name),
				     GUINT_TO_POINTER (the_type));
	}
	g_free (ifaces_quarks);
	g_free (type_name);
	return the_type;
}

MalemTaskDefault *
malem_task_default_new (const gchar *ifaces[],
			const gchar *title,
			const gchar *owner)
{
	GType task_type = malem_task_default_get_type (ifaces);

	return MALEM_TASK_DEFAULT (g_object_new (task_type,
						 "title", title,
						 "owner", owner,
						 NULL));
}
