/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2006 Steve Frécinaux <steve@istique.net>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include "malem-task-list.h"
#include "malem-debug.h"

enum {
	TASK_ADDED,
	TASK_REMOVED,

	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

G_DEFINE_TYPE(MalemTaskList, malem_task_list, G_TYPE_OBJECT);

static void
malem_task_list_init (MalemTaskList *tasklist)
{
	malem_debug ("task-list", "MalemTaskList Initialized");
}

static void
list_object_unref_func (MalemTask     *task,
			MalemTaskList *tasklist)
{
	g_object_unref (task);
}

static void
malem_task_list_finalize (GObject *object)
{
	MalemTaskList *tasklist = MALEM_TASK_LIST (object);
	
	g_list_foreach (tasklist->first,
			(GFunc) list_object_unref_func,
			tasklist);
	g_list_free (tasklist->first);

	malem_debug ("task-list", "MalemTaskList Finalized");

	/* chainup */
	G_OBJECT_CLASS (malem_task_list_parent_class)->finalize (object);
}

static void
malem_task_list_class_init (MalemTaskListClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	GType the_type = G_TYPE_FROM_CLASS (klass);

	object_class->finalize = malem_task_list_finalize;
	
	signals[TASK_ADDED] =
		g_signal_new ("task-added",
			      the_type,
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (MalemTaskListClass, task_added),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOXED,
			      G_TYPE_NONE,
			      1, MALEM_TYPE_TASK_LIST_ITER);

	signals[TASK_REMOVED] =
		g_signal_new ("task-removed",
			      the_type,
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (MalemTaskListClass, task_removed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOXED,
			      G_TYPE_NONE,
			      1, MALEM_TYPE_TASK_LIST_ITER);
}

MalemTaskList *
malem_task_list_new (void)
{
	return MALEM_TASK_LIST (g_object_new (MALEM_TYPE_TASK_LIST, NULL));
}

static inline void
set_list_iter (MalemTaskList     *tasklist,
	       MalemTaskListIter *iter,
	       GList             *llink)
{
	iter->stamp = tasklist->stamp;
	iter->llink = llink;
	iter->tasklist = (gpointer) tasklist;
}

static void
on_task_status_changed (MalemTask       *task,
			MalemTaskStatus  status,
			MalemTaskList   *tasklist)
{
	if (status == MALEM_TASK_STATUS_COMPLETED ||
	    status == MALEM_TASK_STATUS_ABORTED) {
		MalemTaskListIter iter;

		malem_task_list_find (tasklist, &iter, task);
		malem_task_list_remove (tasklist, &iter);
	}
}

void
malem_task_list_add_task (MalemTaskList *tasklist,
			  MalemTask     *task)
{
	MalemTaskListIter iter;
	
	g_return_if_fail (MALEM_IS_TASK_LIST (tasklist));
	g_return_if_fail (MALEM_IS_TASK (task));

	if (tasklist->first == NULL) {
		tasklist->first = g_list_prepend (NULL, task);
		tasklist->last = tasklist->first;
		tasklist->length = 1;
	} else {
		tasklist->last = g_list_append (tasklist->last, task);
		tasklist->last = tasklist->last->next;
		tasklist->length++;
	}

	g_object_ref_sink (task);
	
	g_signal_connect_after (task, "status-changed",
			        G_CALLBACK (on_task_status_changed),
			        tasklist);

	set_list_iter (tasklist, &iter, tasklist->last);
	g_signal_emit (tasklist, signals[TASK_ADDED], 0, &iter);
	
	malem_debug ("task-list", "Task Addded");
}

gboolean
malem_task_list_remove (MalemTaskList     *tasklist,
			MalemTaskListIter *iter)
{	
	g_return_val_if_fail (MALEM_IS_TASK_LIST (tasklist), FALSE);
	g_return_val_if_fail (iter != NULL, FALSE);
	g_return_val_if_fail (iter->stamp == tasklist->stamp, FALSE);

	g_signal_emit (tasklist, signals[TASK_REMOVED], 0, iter);

	g_object_unref (malem_task_list_iter_get_task (iter));

	if (tasklist->last == iter->llink)
		tasklist->last = tasklist->last->prev;
	tasklist->first = g_list_remove_link (tasklist->first, iter->llink);
	tasklist->length--;

	malem_debug ("task-list", "Task Removed");
	return TRUE;
}

void
malem_task_list_find (MalemTaskList     *tasklist,
		      MalemTaskListIter *iter,
		      MalemTask         *task)
{
	GList *llink;

	g_return_if_fail (MALEM_IS_TASK_LIST (tasklist));
	g_return_if_fail (iter != NULL);
	
	llink = g_list_find (tasklist->first, (gconstpointer) task);
	set_list_iter (tasklist, iter, llink);
}

guint
malem_task_list_get_length (MalemTaskList *tasklist)
{
	g_return_val_if_fail (MALEM_IS_TASK_LIST (tasklist), 0);
	return tasklist->length;
}

void
malem_task_list_get_iter_first (MalemTaskList     *tasklist,
				MalemTaskListIter *iter)
{
	g_return_if_fail (MALEM_IS_TASK_LIST (tasklist));
	g_return_if_fail (iter != NULL);
	
	set_list_iter (tasklist, iter, tasklist->first);
}

void
malem_task_list_get_iter_last (MalemTaskList     *tasklist,
			       MalemTaskListIter *iter)
{
	g_return_if_fail (MALEM_IS_TASK_LIST (tasklist));
	g_return_if_fail (iter != NULL);
	
	set_list_iter (tasklist, iter, tasklist->last);
}

void
malem_task_list_get_iter_at_position (MalemTaskList     *tasklist,
				      MalemTaskListIter *iter,
				      guint              pos)
{
	g_return_if_fail (MALEM_IS_TASK_LIST (tasklist));
	g_return_if_fail (iter != NULL);

	set_list_iter (tasklist, iter, g_list_nth (tasklist->first, pos));
}
