#!/usr/bin/env python
# -*- coding: utf-8 -*-
#    Mathusalem Test Application
#    Copyright © 2006  Steve Frécinaux <steve@istique.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import client
import gtk
from gtk import glade

GLADE_FILE = os.path.join(os.path.dirname(__file__), "testapp.glade")

INTERFACES = (
    'org.gnome.Mathusalem.Dictionary',
    'org.gnome.Mathusalem.FileTransfer',
    'org.gnome.Mathusalem.Frobnicator',
)

class MalemTestApp(object):
    TITLE_COLUMN = 0
    PROGRESS_COLUMN = 1
    ICON_COLUMN = 2
    OBJECT_COLUMN = 3
    
    icons = {
        "new": gtk.STOCK_NEW,
        "started": gtk.STOCK_MEDIA_PLAY,
        "paused": gtk.STOCK_MEDIA_PAUSE,
        "aborted": gtk.STOCK_STOP,
        "completed": gtk.STOCK_APPLY,
    }
    
    def __init__(self):
        self.factory = client.Factory()
        self.task = None # Current task being detailed

        callbacks = {
            'on_start_button_clicked'       : self.on_start_button_clicked,
            'on_pause_button_clicked'       : self.on_pause_button_clicked,
            'on_abort_button_clicked'       : self.on_abort_button_clicked,
            'on_new_task_button_clicked'    : self.on_new_task_button_clicked,
    	}

        # Load the "main-window" widget from the glade file.
        self.ui = glade.XML(GLADE_FILE, 'main-window')
        self.ui.signal_autoconnect(callbacks)
        self.window = self['main-window']
        self.window.props.title += ' (version %d.%d.%d)' % self.factory.version
        self.view = self['task-view']

        self.progress_changed_handler = \
            self['task-progress'].connect('value-changed',
                                          self.on_progress_slider_value_changed)

        self.init_task_view()

        self.factory.connect('task-registered', self.on_task_registered)

        ifaces_view = self['task-interfaces'];
        column = gtk.TreeViewColumn('Interface', gtk.CellRendererText(),
                                    text = 0)
        ifaces_view.append_column(column)
        ifaces_view.set_model(gtk.ListStore(str))

        self.update_selected_task()
        self.window.connect('hide', lambda *a: gtk.main_quit())
        self.window.show()

    def init_task_view(self):
        self.model = gtk.ListStore(str, int, str, object)
        for task in self.factory.get_registered_tasks():
            self.add_task(task)
        self.view.set_model(self.model)

        column = gtk.TreeViewColumn('State', gtk.CellRendererPixbuf(),
                                    stock_id = self.ICON_COLUMN)
        self.view.append_column(column)

        column = gtk.TreeViewColumn('Title', gtk.CellRendererText(), 
                                    text = self.TITLE_COLUMN)
        column.set_min_width(275)
        self.view.append_column(column)

        column = gtk.TreeViewColumn('Progress Status', gtk.CellRendererProgress(), 
                                    text = self.PROGRESS_COLUMN,
                                    value = self.PROGRESS_COLUMN)
        column.set_min_width(50)
        self.view.append_column(column)
        self.view.get_selection().connect('changed', self.on_task_view_selection_changed)

    def __getitem__(self, key):
        """Convenience function to get a widget from its name"""
        return self.ui.get_widget(key)

    def add_task(self, task):
        iter = self.model.append([task.title, task.progress,
                                  self.icons[task.status], task])
        task.connect_after('started', self.on_task_started, iter)
        task.connect_after('paused', self.on_task_paused, iter)
        task.connect_after('completed', self.on_task_completed, iter)
        task.connect_after('aborted', self.on_task_aborted, iter)
        task.connect_after('progress-changed', self.on_task_progress_changed, iter)
    
    def on_task_registered(self, factory, task):
        self.add_task(task)
    
    def on_task_started(self, task, iter):
        self.model.set_value(iter, self.ICON_COLUMN, gtk.STOCK_MEDIA_PLAY)
        if task is self.task:
            self.update_selected_task()

    def on_task_paused(self, task, iter):
        self.model.set_value(iter, self.ICON_COLUMN, gtk.STOCK_MEDIA_PAUSE)
        if task is self.task:
            self.update_selected_task()

    def on_task_completed(self, task, iter):
        self.model.set_value(iter, self.ICON_COLUMN, gtk.STOCK_APPLY)
        if task is self.task:
            self.update_selected_task()

    def on_task_aborted(self, task, iter):
        self.model.set_value(iter, self.ICON_COLUMN, gtk.STOCK_STOP)
        if task is self.task:
            self.update_selected_task()
    
    def on_task_progress_changed(self, task, progress, iter):
        task = self.model.get_value(iter, self.OBJECT_COLUMN)
        self.model.set_value(iter, self.PROGRESS_COLUMN, progress)
        if task is self.task:
            self.update_selected_task()
    
    def update_selected_task(self):
        if self.task is not None:       
            title = self.task.title
            owner = self.task.owner
            path = self.task.path
            progress = self.task.progress
            status = self.task.status
        else:
            title = 'None'
            owner = path = status = ''
            progress = 0
        
        title = title.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')        
        self['selected-task-label'].set_markup('<b>Selected Task:</b>  <i>%s</i>' % title)

        self['task-owner'].set_text(owner)
        self['task-object-path'].set_text(path);
        
        ifaces_model = self['task-interfaces'].get_model()
        ifaces_model.clear()
        if self.task is not None:
            for i in self.task.interfaces.iterkeys():
                ifaces_model.append([i])
        
        slider = self['task-progress']
        slider.handler_block(self.progress_changed_handler)
        slider.set_range(0, 110) # small workaround to actually reach 100
        slider.set_value(progress)
        slider.handler_unblock(self.progress_changed_handler)
            
        if status in ['new', 'paused']:
            self['start-button'].set_sensitive(True)
            self['pause-button'].set_sensitive(False)
            self['abort-button'].set_sensitive(True)
            self['task-progress'].set_sensitive(False)
        elif status == 'started':
            self['start-button'].set_sensitive(False)
            self['pause-button'].set_sensitive(True)
            self['abort-button'].set_sensitive(True)
            self['task-progress'].set_sensitive(True)
        else:
            self['start-button'].set_sensitive(False)
            self['pause-button'].set_sensitive(False)
            self['abort-button'].set_sensitive(False)
            self['task-progress'].set_sensitive(False)

    def on_task_view_selection_changed(self, selection):
        model, iter = selection.get_selected()
        
        if iter is not None:
            self.task = model.get_value(iter, self.OBJECT_COLUMN);
        else:
            self.task = None
        self.update_selected_task()

    def on_start_button_clicked(self, button):
        self.task.start()
    
    def on_pause_button_clicked(self, button):
        self.task.pause()
    
    def on_abort_button_clicked(self, button):
        self.task.abort()
    
    def on_progress_slider_value_changed(self, slider):
        self.task.progress = slider.get_value()

    def on_new_task_button_clicked(self, button):
        TaskCreator(self)
    

class TaskCreator(object):
    IFACE_COLUMN = 0
    USE_COLUMN = 1

    def __init__(self, parent):
        self.parent = parent

        callbacks = {
            'on_dialog_response'            : self.on_dialog_response,
		}

        # Load the "main-window" widget from the glade file.
        self.ui = glade.XML(GLADE_FILE, 'task-creator-dialog')
        self.ui.signal_autoconnect(callbacks)
        self.dialog = self['task-creator-dialog']
        self.view = self['interface-view']
        self.init_interfaces_view()
        self.dialog.connect('hide', self.on_dialog_hide)
        self.dialog.show()
    
    def init_interfaces_view(self):
        self.model = gtk.ListStore(str, bool)

        for iface in INTERFACES:
            self.model.append([iface, False])

        self.view.set_model(self.model)

        renderer = gtk.CellRendererToggle()
        column = gtk.TreeViewColumn('Use', renderer, active = self.USE_COLUMN)
        self.view.append_column(column)
        renderer.connect('toggled', self.on_toggle_renderer_toggled)

        column = gtk.TreeViewColumn('Interface', gtk.CellRendererText(), 
                                    text = self.IFACE_COLUMN)
        self.view.append_column(column)

    def __getitem__(self, key):
        """Convenience function to get a widget from its name"""
        return self.ui.get_widget(key)
    
    def on_toggle_renderer_toggled(self, renderer, path):
        iter = self.model.get_iter(path)
        value = self.model.get_value(iter, self.USE_COLUMN)
        self.model.set_value(iter, self.USE_COLUMN, not value)
    
    def on_dialog_hide(self, dialog):
        dialog.destroy()
    
    def on_dialog_response(self, dialog, response_id):
        if response_id == gtk.RESPONSE_OK:
            ifaces = [row[self.IFACE_COLUMN] for row in self.model if row[self.USE_COLUMN] == True]
            title = self['task-title'].get_text()
            owner = self['task-owner'].get_text()
            client.Factory().register_task(ifaces, title, owner)
        dialog.destroy()

if __name__ == '__main__':
    try:
        MalemTestApp()
        gtk.main()
    except client.dbus.DBusException:
        print "Can't connect to the Mathusalem service. Check whether it's running."

# ex:ts=4:et:
